"""
Tencent Cloud Cloud Module
=============================

.. versionadded:: 3000

The Tencent Cloud Cloud Module is used to control access to the Tencent Cloud instance.
https://intl.cloud.tencent.com/

To use this module, set up the cloud configuration at
 ``/etc/salt/cloud.providers`` or ``/etc/salt/cloud.providers.d/*.conf``:

.. code-block:: yaml

    my-tencentcloud-config:
      driver: tencentcloud
      # Tencent Cloud Secret Id
      id: AKIDA64pOio9BMemkApzevX0HS169S4b750A
      # Tencent Cloud Secret Key
      key: 8r2xmPn0C5FDvRAlmcJimiTZKVRsk260
      # Tencent Cloud Region
      location: ap-guangzhou

:depends: tencentcloud-sdk-python
"""

import logging
import pprint
import time

import salt.config as config
import salt.utils.cloud
import salt.utils.data
import salt.utils.json
from salt.exceptions import (
    SaltCloudExecutionFailure,
    SaltCloudExecutionTimeout,
    SaltCloudNotFound,
    SaltCloudSystemExit,
)

try:
    # Try import tencentcloud sdk
    from tencentcloud.common import credential  # pylint: disable=no-name-in-module

    # pylint: disable=no-name-in-module
    from tencentcloud.common.profile.client_profile import ClientProfile
    from tencentcloud.cvm.v20170312 import cvm_client
    from tencentcloud.cvm.v20170312 import models as cvm_models
    from tencentcloud.vpc.v20170312 import models as vpc_models
    from tencentcloud.vpc.v20170312 import vpc_client

    # pylint: enable=no-name-in-module

    HAS_TENCENTCLOUD_SDK = True
except ImportError:
    HAS_TENCENTCLOUD_SDK = False

# Get logging started
log = logging.getLogger(__name__)

# The default region
DEFAULT_REGION = "ap-guangzhou"

# The Tencent Cloud
__virtualname__ = "tencentcloud"


def __virtual__():
    """
    Only load in this module if the Tencent Cloud configurations are in place
    """
    if get_configured_provider() is False:
        return False

    if get_dependencies() is False:
        return False

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__, _get_active_provider_name() or __virtualname__, ("id", "key")
    )


def get_dependencies():
    """
    Warn if dependencies aren't met.
    """
    return config.check_driver_dependencies(
        __virtualname__, {"tencentcloud-sdk-python": HAS_TENCENTCLOUD_SDK}
    )


def get_provider_client(name=None):
    """
    Return a new provider client
    """
    provider = get_configured_provider()

    secretId = provider.get("id")
    secretKey = provider.get("key")
    region = __get_location(None)

    cpf = ClientProfile()
    cpf.language = "en-US"
    crd = credential.Credential(secretId, secretKey)

    if name == "cvm_client":
        client = cvm_client.CvmClient(crd, region, cpf)
    elif name == "vpc_client":
        client = vpc_client.VpcClient(crd, region, cpf)
    else:
        raise SaltCloudSystemExit("Client name {} is not supported".format(name))

    return client


def avail_locations(call=None):
    """
    Return Tencent Cloud available region

    CLI Example:

    .. code-block:: bash

        salt-cloud --list-locations my-tencentcloud-config
        salt-cloud -f avail_locations my-tencentcloud-config
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_locations function must be called with "
            "-f or --function, or with the --list-locations option"
        )

    client = get_provider_client("cvm_client")
    req = cvm_models.DescribeRegionsRequest()
    resp = client.DescribeRegions(req)

    ret = {}
    for region in resp.RegionSet:
        if region.RegionState != "AVAILABLE":
            continue
        ret[region.Region] = region.RegionName

    return ret


def avail_images(call=None):
    """
    Return Tencent Cloud available image

    CLI Example:

    .. code-block:: bash

        salt-cloud --list-images my-tencentcloud-config
        salt-cloud -f avail_images my-tencentcloud-config
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with "
            "-f or --function, or with the --list-images option"
        )

    return _get_images(
        ["PUBLIC_IMAGE", "PRIVATE_IMAGE", "IMPORT_IMAGE", "SHARED_IMAGE"]
    )


def avail_sizes(call=None):
    """
    Return Tencent Cloud available instance type

    CLI Example:

    .. code-block:: bash

        salt-cloud --list-sizes my-tencentcloud-config
        salt-cloud -f avail_sizes my-tencentcloud-config
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with "
            "-f or --function, or with the --list-sizes option"
        )

    client = get_provider_client("cvm_client")
    req = cvm_models.DescribeInstanceTypeConfigsRequest()
    resp = client.DescribeInstanceTypeConfigs(req)

    ret = {}
    for typeConfig in resp.InstanceTypeConfigSet:
        ret[typeConfig.InstanceType] = {
            "Zone": typeConfig.Zone,
            "InstanceFamily": typeConfig.InstanceFamily,
            "Memory": "{}GB".format(typeConfig.Memory),
            "CPU": "{}-Core".format(typeConfig.CPU),
        }
        if typeConfig.GPU:
            ret[typeConfig.InstanceType]["GPU"] = "{}-Core".format(typeConfig.GPU)

    return ret


def list_securitygroups(call=None):
    """
    Return all Tencent Cloud security groups in current region

    CLI Example:

    .. code-block:: bash

        salt-cloud -f list_securitygroups my-tencentcloud-config
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_securitygroups function must be called with -f or --function."
        )

    client = get_provider_client("vpc_client")
    req = vpc_models.DescribeSecurityGroupsRequest()
    req.Offset = 0
    req.Limit = 100
    resp = client.DescribeSecurityGroups(req)

    ret = {}
    for sg in resp.SecurityGroupSet:
        ret[sg.SecurityGroupId] = {
            "SecurityGroupName": sg.SecurityGroupName,
            "SecurityGroupDesc": sg.SecurityGroupDesc,
            "ProjectId": sg.ProjectId,
            "IsDefault": sg.IsDefault,
            "CreatedTime": sg.CreatedTime,
        }

    return ret


def list_custom_images(call=None):
    """
    Return all Tencent Cloud images in current region

    CLI Example:

    .. code-block:: bash

        salt-cloud -f list_custom_images my-tencentcloud-config
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_custom_images function must be called with -f or --function."
        )

    return _get_images(["PRIVATE_IMAGE", "IMPORT_IMAGE"])


def list_availability_zones(call=None):
    """
    Return all Tencent Cloud availability zones in current region

    CLI Example:

    .. code-block:: bash

        salt-cloud -f list_availability_zones my-tencentcloud-config
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_availability_zones function must be called with -f or --function."
        )

    client = get_provider_client("cvm_client")
    req = cvm_models.DescribeZonesRequest()
    resp = client.DescribeZones(req)

    ret = {}
    for zone in resp.ZoneSet:
        if zone.ZoneState != "AVAILABLE":
            continue
        ret[zone.Zone] = (zone.ZoneName,)

    return ret


def list_nodes(call=None):
    """
    Return a list of instances that are on the provider

    CLI Examples:

    .. code-block:: bash

        salt-cloud -Q
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes function must be called with -f or --function."
        )

    ret = {}
    nodes = _get_nodes()
    for instance in nodes:
        ret[instance.InstanceId] = {
            "InstanceId": instance.InstanceId,
            "InstanceName": instance.InstanceName,
            "InstanceType": instance.InstanceType,
            "ImageId": instance.ImageId,
            "PublicIpAddresses": instance.PublicIpAddresses,
            "PrivateIpAddresses": instance.PrivateIpAddresses,
            "InstanceState": instance.InstanceState,
        }

    return ret


def list_nodes_full(call=None):
    """
    Return a list of instances that are on the provider, with full details

    CLI Examples:

    .. code-block:: bash

        salt-cloud -F
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_full function must be called with -f or --function."
        )

    ret = {}
    nodes = _get_nodes()
    for instance in nodes:
        instanceAttribute = vars(instance)
        ret[instance.InstanceName] = instanceAttribute
        for k in [
            "DataDisks",
            "InternetAccessible",
            "LoginSettings",
            "Placement",
            "SystemDisk",
            "Tags",
            "VirtualPrivateCloud",
        ]:
            ret[instance.InstanceName][k] = str(instanceAttribute[k])

    provider = _get_active_provider_name() or "tencentcloud"
    if ":" in provider:
        comps = provider.split(":")
        provider = comps[0]

    __opts__["update_cachedir"] = True
    __utils__["cloud.cache_node_list"](ret, provider, __opts__)

    return ret


def list_nodes_select(call=None):
    """
    Return a list of instances that are on the provider, with select fields

    CLI Examples:

    .. code-block:: bash

        salt-cloud -S
    """
    return salt.utils.cloud.list_nodes_select(
        list_nodes_full("function"),
        __opts__["query.selection"],
        call,
    )


def list_nodes_min(call=None):
    """
    Return a list of instances that are on the provider, Only names, and their state, is returned.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f list_nodes_min my-tencentcloud-config
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_min function must be called with -f or --function."
        )

    ret = {}
    nodes = _get_nodes()
    for instance in nodes:
        ret[instance.InstanceName] = {
            "InstanceId": instance.InstanceId,
            "InstanceState": instance.InstanceState,
        }

    return ret


def create(vm_):
    """
    Create a single Tencent Cloud instance from a data dict.

    Tencent Cloud profiles require a ``provider``, ``availability_zone``, ``image`` and ``size``.
    Set up profile at ``/etc/salt/cloud.profiles`` or ``/etc/salt/cloud.profiles.d/*.conf``:

    .. code-block:: yaml

        tencentcloud-guangzhou-s1sm1:
            provider: my-tencentcloud-config
            availability_zone: ap-guangzhou-3
            image: img-31tjrtph
            size: S1.SMALL1
            allocate_public_ip: True
            internet_max_bandwidth_out: 1
            password: '153e41ec96140152'
            securitygroups:
                - sg-5e90804b

    CLI Examples:

    .. code-block:: bash

        salt-cloud -p tencentcloud-guangzhou-s1 myinstance
    """
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__,
                _get_active_provider_name() or "tencentcloud",
                vm_["profile"],
                vm_=vm_,
            )
            is False
        ):
            return False
    except AttributeError:
        pass

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{}/creating".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "creating", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    log.debug("Try creating instance: %s", pprint.pformat(vm_))

    # Init cvm client
    client = get_provider_client("cvm_client")
    req = cvm_models.RunInstancesRequest()
    req.InstanceName = vm_["name"]

    # Required parameters
    req.InstanceType = __get_size(vm_)
    req.ImageId = __get_image(vm_)

    zone = __get_availability_zone(vm_)
    projectId = vm_.get("project_id", 0)
    req.Placement = {"Zone": zone, "ProjectId": projectId}

    # Optional parameters

    req.SecurityGroupIds = __get_securitygroups(vm_)
    req.HostName = vm_.get("hostname", vm_["name"])

    req.InstanceChargeType = vm_.get("instance_charge_type", "POSTPAID_BY_HOUR")
    if req.InstanceChargeType == "PREPAID":
        period = vm_.get("instance_charge_type_prepaid_period", 1)
        renewFlag = vm_.get(
            "instance_charge_type_prepaid_renew_flag", "NOTIFY_AND_MANUAL_RENEW"
        )
        req.InstanceChargePrepaid = {"Period": period, "RenewFlag": renewFlag}

    allocate_public_ip = vm_.get("allocate_public_ip", False)
    internet_max_bandwidth_out = vm_.get("internet_max_bandwidth_out", 0)
    if allocate_public_ip and internet_max_bandwidth_out > 0:
        req.InternetAccessible = {
            "PublicIpAssigned": allocate_public_ip,
            "InternetMaxBandwidthOut": internet_max_bandwidth_out,
        }
        internet_charge_type = vm_.get("internet_charge_type", "")
        if internet_charge_type != "":
            req.InternetAccessible["InternetChargeType"] = internet_charge_type

    req.LoginSettings = {}
    req.VirtualPrivateCloud = {}
    req.SystemDisk = {}

    keyId = vm_.get("key_name", "")
    if keyId:
        req.LoginSettings["KeyIds"] = [keyId]

    password = vm_.get("password", "")
    if password:
        req.LoginSettings["Password"] = password

    private_ip = vm_.get("private_ip", "")
    if private_ip:
        req.VirtualPrivateCloud["PrivateIpAddresses"] = private_ip

    vpc_id = vm_.get("vpc_id", "")
    if vpc_id:
        req.VirtualPrivateCloud["VpcId"] = vpc_id

    subnetId = vm_.get("subnet_id", "")
    if subnetId:
        req.VirtualPrivateCloud["SubnetId"] = subnetId

    system_disk_size = vm_.get("system_disk_size", 0)
    if system_disk_size:
        req.SystemDisk["DiskSize"] = system_disk_size

    system_disk_type = vm_.get("system_disk_type", "")
    if system_disk_type:
        req.SystemDisk["DiskType"] = system_disk_type

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(vm_["name"]),
        args=__utils__["cloud.filter_event"]("requesting", vm_, list(vm_)),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    try:
        resp = client.RunInstances(req)
        if not resp.InstanceIdSet:
            raise SaltCloudSystemExit("Unexpected error, no instance created")
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            "Error creating %s on tencentcloud\n\n"
            "The following exception was thrown when trying to "
            "run the initial deployment: %s",
            vm_["name"],
            str(exc),
            # Show the traceback if the debug logging level is enabled
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False

    time.sleep(5)

    def __query_node_data(vm_name):
        data = show_instance(vm_name, call="action")
        if not data:
            return False
        if data["InstanceState"] != "RUNNING":
            return False
        if data["PrivateIpAddresses"]:
            return data

    try:
        data = salt.utils.cloud.wait_for_ip(
            __query_node_data,
            update_args=(vm_["name"],),
            timeout=config.get_cloud_config_value(
                "wait_for_ip_timeout", vm_, __opts__, default=10 * 60
            ),
            interval=config.get_cloud_config_value(
                "wait_for_ip_interval", vm_, __opts__, default=10
            ),
        )
    except (SaltCloudExecutionTimeout, SaltCloudExecutionFailure) as exc:
        try:
            destroy(vm_["name"])
        except SaltCloudSystemExit:
            pass
        finally:
            raise SaltCloudSystemExit(str(exc))

    if data["PublicIpAddresses"]:
        ssh_ip = data["PublicIpAddresses"][0]
    elif data["PrivateIpAddresses"]:
        ssh_ip = data["PrivateIpAddresses"][0]
    else:
        log.error("No available ip: cant connect to salt")
        return False

    log.debug("Instance %s: %s is now running", vm_["name"], ssh_ip)
    vm_["ssh_host"] = ssh_ip

    # The instance is booted and accessible, let's Salt it!
    ret = __utils__["cloud.bootstrap"](vm_, __opts__)
    ret.update(data)

    log.debug("'%s' instance creation details:\n%s", vm_["name"], pprint.pformat(data))

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "created", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return ret


def start(name, call=None):
    """
    Start a Tencent Cloud instance
    Notice: the instance state must be stopped

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a start myinstance
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    node = _get_node(name)

    client = get_provider_client("cvm_client")
    req = cvm_models.StartInstancesRequest()
    req.InstanceIds = [node.InstanceId]
    resp = client.StartInstances(req)

    return resp


def stop(name, force=False, call=None):
    """
    Stop a Tencent Cloud running instance
    Note: use `force=True` to make force stop

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a stop myinstance
        salt-cloud -a stop myinstance force=True
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    node = _get_node(name)

    client = get_provider_client("cvm_client")
    req = cvm_models.StopInstancesRequest()
    req.InstanceIds = [node.InstanceId]
    if force:
        req.ForceStop = "TRUE"
    resp = client.StopInstances(req)

    return resp


def reboot(name, call=None):
    """
    Reboot a Tencent Cloud instance

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a reboot myinstance
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    node = _get_node(name)

    client = get_provider_client("cvm_client")
    req = cvm_models.RebootInstancesRequest()
    req.InstanceIds = [node.InstanceId]
    resp = client.RebootInstances(req)

    return resp


def destroy(name, call=None):
    """
    Destroy a Tencent Cloud instance

    CLI Example:

    .. code-block:: bash

        salt-cloud -a destroy myinstance
        salt-cloud -d myinstance
    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        "salt/cloud/{}/destroying".format(name),
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    node = _get_node(name)

    client = get_provider_client("cvm_client")
    req = cvm_models.TerminateInstancesRequest()
    req.InstanceIds = [node.InstanceId]
    resp = client.TerminateInstances(req)

    __utils__["cloud.fire_event"](
        "event",
        "destroyed instance",
        "salt/cloud/{}/destroyed".format(name),
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return resp


def script(vm_):
    """
    Return the script deployment object
    """
    return salt.utils.cloud.os_script(
        config.get_cloud_config_value("script", vm_, __opts__),
        vm_,
        __opts__,
        salt.utils.cloud.salt_config_to_yaml(
            salt.utils.cloud.minion_config(__opts__, vm_)
        ),
    )


def show_image(kwargs, call=None):
    """
    Show the details of Tencent Cloud image

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f show_image tencentcloud image=img-31tjrtph
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "The show_image function must be called with -f or --function"
        )

    if not isinstance(kwargs, dict):
        kwargs = {}

    if "image" not in kwargs:
        raise SaltCloudSystemExit("No image specified.")

    image = kwargs["image"]

    client = get_provider_client("cvm_client")
    req = cvm_models.DescribeImagesRequest()
    req.ImageIds = [image]
    resp = client.DescribeImages(req)

    if not resp.ImageSet:
        raise SaltCloudNotFound(
            "The specified image '{}' could not be found.".format(image)
        )

    ret = {}
    for image in resp.ImageSet:
        ret[image.ImageId] = {
            "ImageName": image.ImageName,
            "ImageType": image.ImageType,
            "ImageSource": image.ImageSource,
            "Platform": image.Platform,
            "Architecture": image.Architecture,
            "ImageSize": "{}GB".format(image.ImageSize),
            "ImageState": image.ImageState,
        }

    return ret


def show_instance(name, call=None):
    """
    Show the details of Tencent Cloud instance

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a show_instance myinstance
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_instance action must be called with -a or --action."
        )

    node = _get_node(name)
    ret = vars(node)
    for k in [
        "DataDisks",
        "InternetAccessible",
        "LoginSettings",
        "Placement",
        "SystemDisk",
        "Tags",
        "VirtualPrivateCloud",
    ]:
        ret[k] = str(ret[k])

    return ret


def show_disk(name, call=None):
    """
    Show the disk details of Tencent Cloud instance

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a show_disk myinstance
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_disks action must be called with -a or --action."
        )

    node = _get_node(name)

    ret = {}
    ret[node.SystemDisk.DiskId] = {
        "SystemDisk": True,
        "DiskSize": node.SystemDisk.DiskSize,
        "DiskType": node.SystemDisk.DiskType,
        "DeleteWithInstance": True,
        "SnapshotId": "",
    }

    if node.DataDisks:
        for disk in node.DataDisks:
            ret[disk.DiskId] = {
                "SystemDisk": False,
                "DiskSize": disk.DiskSize,
                "DiskType": disk.DiskType,
                "DeleteWithInstance": disk.DeleteWithInstance,
                "SnapshotId": disk.SnapshotId,
            }

    return ret


def _get_node(name):
    """
    Return Tencent Cloud instance detail by name
    """
    attempts = 5
    while attempts >= 0:
        try:
            client = get_provider_client("cvm_client")
            req = cvm_models.DescribeInstancesRequest()
            req.Filters = [{"Name": "instance-name", "Values": [name]}]
            resp = client.DescribeInstances(req)
            return resp.InstanceSet[0]
        except Exception as ex:  # pylint: disable=broad-except
            attempts -= 1
            log.debug(
                "Failed to get data for node '%s': %s. Remaining attempts: %d",
                name,
                ex,
                attempts,
            )
            time.sleep(0.5)

    raise SaltCloudNotFound("Failed to get instance info {}".format(name))


def _get_nodes():
    """
    Return all list of Tencent Cloud instances
    """
    ret = []
    offset = 0
    limit = 100

    while True:
        client = get_provider_client("cvm_client")
        req = cvm_models.DescribeInstancesRequest()
        req.Offset = offset
        req.Limit = limit
        resp = client.DescribeInstances(req)
        for v in resp.InstanceSet:
            ret.append(v)
        if len(ret) >= resp.TotalCount:
            break
        offset += len(resp.InstanceSet)

    return ret


def _get_images(image_type):
    """
    Return all list of Tencent Cloud images
    """
    client = get_provider_client("cvm_client")
    req = cvm_models.DescribeImagesRequest()
    req.Filters = [{"Name": "image-type", "Values": image_type}]
    req.Offset = 0
    req.Limit = 100
    resp = client.DescribeImages(req)

    ret = {}
    for image in resp.ImageSet:
        if image.ImageState != "NORMAL":
            continue
        ret[image.ImageId] = {
            "ImageName": image.ImageName,
            "ImageType": image.ImageType,
            "ImageSource": image.ImageSource,
            "Platform": image.Platform,
            "Architecture": image.Architecture,
            "ImageSize": "{}GB".format(image.ImageSize),
        }

    return ret


def __get_image(vm_):
    vm_image = str(
        config.get_cloud_config_value("image", vm_, __opts__, search_global=False)
    )

    if not vm_image:
        raise SaltCloudNotFound("No image specified.")

    images = avail_images()
    if vm_image in images:
        return vm_image

    raise SaltCloudNotFound(
        "The specified image '{}' could not be found.".format(vm_image)
    )


def __get_size(vm_):
    vm_size = str(
        config.get_cloud_config_value("size", vm_, __opts__, search_global=False)
    )

    if not vm_size:
        raise SaltCloudNotFound("No size specified.")

    sizes = avail_sizes()
    if vm_size in sizes:
        return vm_size

    raise SaltCloudNotFound(
        "The specified size '{}' could not be found.".format(vm_size)
    )


def __get_securitygroups(vm_):
    vm_securitygroups = config.get_cloud_config_value(
        "securitygroups", vm_, __opts__, search_global=False
    )

    if not vm_securitygroups:
        return []

    securitygroups = list_securitygroups()
    for idx, value in enumerate(vm_securitygroups):
        vm_securitygroups[idx] = str(value)
        if vm_securitygroups[idx] not in securitygroups:
            raise SaltCloudNotFound(
                "The specified securitygroups '{}' could not be found.".format(
                    vm_securitygroups[idx]
                )
            )

    return vm_securitygroups


def __get_availability_zone(vm_):
    vm_availability_zone = str(
        config.get_cloud_config_value(
            "availability_zone", vm_, __opts__, search_global=False
        )
    )

    if not vm_availability_zone:
        raise SaltCloudNotFound("No availability_zone specified.")

    availability_zones = list_availability_zones()
    if vm_availability_zone in availability_zones:
        return vm_availability_zone

    raise SaltCloudNotFound(
        "The specified availability_zone '{}' could not be found.".format(
            vm_availability_zone
        )
    )


def __get_location(vm_):
    """
    Return the Tencent Cloud region to use, in this order:
        - CLI parameter
        - VM parameter
        - Cloud profile setting
    """
    vm_location = str(
        __opts__.get(
            "location",
            config.get_cloud_config_value(
                "location",
                vm_ or get_configured_provider(),
                __opts__,
                default=DEFAULT_REGION,
                search_global=False,
            ),
        )
    )

    if not vm_location:
        raise SaltCloudNotFound("No location specified.")

    return vm_location



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    Test cases for salt.modules.hg
"""


import pytest

import salt.modules.hg as hg
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {hg: {}}


def test_revision():
    """
    Test for Returns the long hash of a given identifier
    """
    mock = MagicMock(
        side_effect=[{"retcode": 0, "stdout": "A"}, {"retcode": 1, "stdout": "A"}]
    )
    with patch.dict(hg.__salt__, {"cmd.run_all": mock}):
        assert hg.revision("cwd") == "A"

        assert hg.revision("cwd") == ""


def test_describe():
    """
    Test for Mimic git describe.
    """
    with patch.dict(hg.__salt__, {"cmd.run_stdout": MagicMock(return_value="A")}):
        with patch.object(hg, "revision", return_value=False):
            assert hg.describe("cwd") == "A"


def test_archive():
    """
    Test for Export a tarball from the repository
    """
    with patch.dict(hg.__salt__, {"cmd.run": MagicMock(return_value="A")}):
        assert hg.archive("cwd", "output") == "A"


def test_pull():
    """
    Test for Perform a pull on the given repository
    """
    with patch.dict(
        hg.__salt__,
        {"cmd.run_all": MagicMock(return_value={"retcode": 0, "stdout": "A"})},
    ):
        assert hg.pull("cwd") == "A"


def test_update():
    """
    Test for Update to a given revision
    """
    with patch.dict(
        hg.__salt__,
        {"cmd.run_all": MagicMock(return_value={"retcode": 0, "stdout": "A"})},
    ):
        assert hg.update("cwd", "rev") == "A"


def test_clone():
    """
    Test for Clone a new repository
    """
    with patch.dict(
        hg.__salt__,
        {"cmd.run_all": MagicMock(return_value={"retcode": 0, "stdout": "A"})},
    ):
        assert hg.clone("cwd", "repository") == "A"


def test_status_single():
    """
    Test for Status to a given repository
    """
    with patch.dict(
        hg.__salt__,
        {"cmd.run_stdout": MagicMock(return_value="A added 0\nA added 1\nM modified")},
    ):
        assert hg.status("cwd") == {
            "added": ["added 0", "added 1"],
            "modified": ["modified"],
        }


def test_status_multiple():
    """
    Test for Status to a given repository (cwd is list)
    """
    with patch.dict(
        hg.__salt__,
        {
            "cmd.run_stdout": MagicMock(
                side_effect=(
                    lambda *args, **kwargs: {
                        "dir 0": "A file 0\n",
                        "dir 1": "M file 1",
                    }[kwargs["cwd"]]
                )
            )
        },
    ):
        assert hg.status(["dir 0", "dir 1"]) == {
            "dir 0": {"added": ["file 0"]},
            "dir 1": {"modified": ["file 1"]},
        }



import tarfile

import pytest

import salt.utils.http


@pytest.mark.parametrize("backend", ["requests", "urllib2", "tornado"])
def test_decode_body(webserver, integration_files_dir, backend):
    with tarfile.open(integration_files_dir / "test.tar.gz", "w:gz") as tar:
        tar.add(integration_files_dir / "this.txt")

    ret = salt.utils.http.query(
        webserver.url("test.tar.gz"), backend=backend, decode_body=False
    )
    assert isinstance(ret["body"], bytes)



"""
    :codeauthor: Alexander Schwartz <alexander.schwartz@gmx.net>
"""
import pytest

import salt.client
from salt.cloud.clouds import saltify
from tests.support.mock import ANY, MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    test_profiles = {
        "testprofile1": NotImplemented,
        "testprofile2": {  # this profile is used in test_saltify_destroy()
            "ssh_username": "fred",
            "remove_config_on_destroy": False,  # expected for test
            "shutdown_on_destroy": True,  # expected value for test
        },
        "testprofile3": {  # this profile is used in test_create_wake_on_lan()
            "wake_on_lan_mac": "aa-bb-cc-dd-ee-ff",
            "wol_sender_node": "friend1",
            "wol_boot_wait": 0.01,  # we want the wait to be very short
        },
    }

    return {
        saltify: {
            "__active_provider_name__": "",
            "__utils__": {
                "cloud.bootstrap": MagicMock(),
                "cloud.fire_event": MagicMock(),
            },
            "__opts__": {
                "providers": {
                    "sfy1": {
                        "saltify": {"driver": "saltify", "profiles": test_profiles}
                    },
                },
                "profiles": test_profiles,
                "sock_dir": "/var/sockxxx",
                "transport": "tcp",
            },
        }
    }


def test_create_no_deploy():
    """
    Test if deployment fails. This is the most basic test as saltify doesn't contain much logic
    """
    with patch("salt.cloud.clouds.saltify._verify", MagicMock(return_value=True)):
        vm = {"deploy": False, "driver": "saltify", "name": "dummy"}
        assert saltify.create(vm)


def test_create_and_deploy():
    """
    Test if deployment can be done.
    """
    mock_cmd = MagicMock(return_value=True)
    with patch.dict(
        "salt.cloud.clouds.saltify.__utils__", {"cloud.bootstrap": mock_cmd}
    ):
        vm_ = {
            "deploy": True,
            "driver": "saltify",
            "name": "new2",
            "profile": "testprofile2",
        }
        result = saltify.create(vm_)
        mock_cmd.assert_called_once_with(vm_, ANY)
        assert result


def test_create_no_ssh_host():
    """
    Test that ssh_host is set to the vm name if not defined
    """
    mock_cmd = MagicMock(return_value=True)
    with patch.dict(
        "salt.cloud.clouds.saltify.__utils__", {"cloud.bootstrap": mock_cmd}
    ):
        vm_ = {
            "deploy": True,
            "driver": "saltify",
            "name": "new2",
            "profile": "testprofile2",
        }
        result = saltify.create(vm_)
        mock_cmd.assert_called_once_with(vm_, ANY)
        assert result
        # Make sure that ssh_host was added to the vm. Note that this is
        # done in two asserts so that the failure is more explicit about
        # what is wrong. If ssh_host wasn't inserted in the vm_ dict, the
        # failure would be a KeyError, which would be harder to
        # troubleshoot.
        assert "ssh_host" in vm_
        assert vm_["ssh_host"] == "new2"


def test_create_wake_on_lan():
    """
    Test if wake on lan works
    """
    mock_sleep = MagicMock()
    mock_cmd = MagicMock(return_value=True)
    mm_cmd = MagicMock(return_value={"friend1": True})
    with salt.client.LocalClient() as lcl:
        lcl.cmd = mm_cmd
        with patch("time.sleep", mock_sleep):
            with patch("salt.client.LocalClient", return_value=lcl):
                with patch.dict(
                    "salt.cloud.clouds.saltify.__utils__",
                    {"cloud.bootstrap": mock_cmd},
                ):
                    vm_ = {
                        "deploy": True,
                        "driver": "saltify",
                        "name": "new1",
                        "profile": "testprofile3",
                    }
                    result = saltify.create(vm_)
                    mock_cmd.assert_called_once_with(vm_, ANY)
                    mm_cmd.assert_called_with(
                        "friend1", "network.wol", ["aa-bb-cc-dd-ee-ff"]
                    )
                    # The test suite might call time.sleep, look for any call
                    # that has the expected wait time.
                    mock_sleep.assert_any_call(0.01)
                    assert result


def test_avail_locations():
    """
    Test the avail_locations will always return {}
    """
    assert saltify.avail_locations() == {}


def test_avail_sizes():
    """
    Test the avail_sizes will always return {}
    """
    assert saltify.avail_sizes() == {}


def test_avail_images():
    """
    Test the avail_images will return profiles
    """
    testlist = list(saltify.__opts__["profiles"].keys())
    assert saltify.avail_images()["Profiles"].sort() == testlist.sort()


def test_list_nodes():
    """
    Test list_nodes will return required fields only
    """
    testgrains = {
        "nodeX1": {
            "id": "nodeX1",
            "ipv4": ["127.0.0.1", "192.1.2.22", "172.16.17.18"],
            "ipv6": ["::1", "fdef:bad:add::f00", "3001:DB8::F00D"],
            "salt-cloud": {
                "driver": "saltify",
                "provider": "saltyfy",
                "profile": "testprofile2",
            },
            "extra_stuff": "does not belong",
        }
    }
    expected_result = {
        "nodeX1": {
            "id": "nodeX1",
            "image": "testprofile2",
            "private_ips": ["172.16.17.18", "fdef:bad:add::f00"],
            "public_ips": ["192.1.2.22", "3001:DB8::F00D"],
            "size": "",
            "state": "running",
        }
    }
    mm_cmd = MagicMock(return_value=testgrains)
    with salt.client.LocalClient() as lcl:
        lcl.cmd = mm_cmd
        with patch("salt.client.LocalClient", return_value=lcl):
            assert saltify.list_nodes() == expected_result


def test_saltify_reboot():
    mm_cmd = MagicMock(return_value=True)
    with salt.client.LocalClient() as lcl:
        lcl.cmd = mm_cmd
        with patch("salt.client.LocalClient", return_value=lcl):
            result = saltify.reboot("nodeS1", "action")
            mm_cmd.assert_called_with("nodeS1", "system.reboot")
            assert result


def test_saltify_destroy():
    # destroy calls local.cmd several times and expects
    # different results, so we will provide a list of
    # results. Each call will get the next value.
    # NOTE: this assumes that the call order never changes,
    # so to keep things simple, we will not use remove_config...
    result_list = [
        {
            "nodeS1": {  # first call is grains.get
                "driver": "saltify",
                "provider": "saltify",
                "profile": "testprofile2",
            }
        },
        #  Note:
        #    testprofile2 has remove_config_on_destroy: False
        #    and shutdown_on_destroy: True
        {
            "nodeS1": (  # last call shuts down the minion
                "a system.shutdown worked message"
            )
        },
    ]
    mm_cmd = MagicMock(side_effect=result_list)
    with salt.client.LocalClient() as lcl:
        lcl.cmd = mm_cmd
        with patch("salt.client.LocalClient", return_value=lcl):
            result = saltify.destroy("nodeS1", "action")
            mm_cmd.assert_called_with("nodeS1", "system.shutdown")
            assert result



import logging

log = logging.getLogger(__name__)


def test_run_job(schedule):
    """
    verify that scheduled job runs
    """
    job_name = "test_run_job"
    job = {"schedule": {job_name: {"function": "test.ping"}}}
    # Add the job to the scheduler
    schedule.opts.update(job)

    # Run job
    schedule.run_job(job_name)
    ret = schedule.job_status(job_name)
    expected = {"function": "test.ping", "run": True, "name": "test_run_job"}
    assert ret == expected



import random

import pytest

import salt.utils.platform
from tests.support.case import ModuleCase


@pytest.mark.windows_whitelisted
class StatusModuleTest(ModuleCase):
    """
    Test the status module
    """

    @pytest.mark.skip_on_windows
    @pytest.mark.flaky(max_runs=4)
    @pytest.mark.slow_test
    def test_status_pid(self):
        """
        status.pid
        """
        status_pid = self.run_function("status.pid", ["salt"])
        grab_pids = status_pid.split()[:10]
        random_pid = random.choice(grab_pids)
        grep_salt = self.run_function("cmd.run", ["pgrep -f salt"])
        self.assertIn(random_pid, grep_salt)

    @pytest.mark.skip_unless_on_windows
    @pytest.mark.slow_test
    def test_status_cpuload(self):
        """
        status.cpuload
        """
        ret = self.run_function("status.cpuload")
        self.assertTrue(isinstance(ret, float))

    @pytest.mark.skip_unless_on_windows
    @pytest.mark.slow_test
    def test_status_saltmem(self):
        """
        status.saltmem
        """
        ret = self.run_function("status.saltmem")
        self.assertTrue(isinstance(ret, int))

    @pytest.mark.slow_test
    def test_status_diskusage(self):
        """
        status.diskusage
        """
        ret = self.run_function("status.diskusage")
        if salt.utils.platform.is_darwin():
            self.assertIn("not yet supported on this platform", ret)
        elif salt.utils.platform.is_windows():
            self.assertTrue(isinstance(ret["percent"], float))
        else:
            self.assertIn("total", str(ret))
            self.assertIn("available", str(ret))

    @pytest.mark.slow_test
    def test_status_procs(self):
        """
        status.procs
        """
        ret = self.run_function("status.procs")
        for x, y in ret.items():
            self.assertIn("cmd", y)

    @pytest.mark.slow_test
    def test_status_uptime(self):
        """
        status.uptime
        """
        ret = self.run_function("status.uptime")

        if salt.utils.platform.is_windows():
            self.assertTrue(isinstance(ret, float))
        else:
            self.assertTrue(isinstance(ret["days"], int))



"""
Management of Zabbix Valuemap object over Zabbix API.

.. versionadded:: 2017.7.0

:codeauthor: Jakub Sliva <jakub.sliva@ultimum.io>
"""

import json
import logging

from salt.exceptions import SaltException

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only make these states available if Zabbix module and run_query function is available
    and all 3rd party modules imported.
    """
    if "zabbix.run_query" in __salt__:
        return True
    return False, "Import zabbix or other needed modules failed."


def present(name, params, **kwargs):
    """
    Creates Zabbix Value map object or if differs update it according defined parameters

    :param name: Zabbix Value map name
    :param params: Definition of the Zabbix Value map
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        zabbix-valuemap-present:
            zabbix_valuemap.present:
                - name: Number mapping
                - params:
                    mappings:
                        - value: 1
                          newvalue: one
                        - value: 2
                          newvalue: two
    """
    zabbix_id_mapper = __salt__["zabbix.get_zabbix_id_mapper"]()

    dry_run = __opts__["test"]
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    # Create input params substituting functions with their results
    params["name"] = name
    input_params = __salt__["zabbix.substitute_params"](params, **kwargs)
    log.info(
        "Zabbix Value map: input params: %s",
        str(json.dumps(input_params, indent=4)),
    )

    search = {"output": "extend", "selectMappings": "extend", "filter": {"name": name}}
    # GET Value map object if exists
    valuemap_get = __salt__["zabbix.run_query"]("valuemap.get", search, **kwargs)
    log.info(
        "Zabbix Value map: valuemap.get result: %s",
        str(json.dumps(valuemap_get, indent=4)),
    )

    existing_obj = (
        __salt__["zabbix.substitute_params"](valuemap_get[0], **kwargs)
        if valuemap_get and len(valuemap_get) == 1
        else False
    )

    if existing_obj:
        diff_params = __salt__["zabbix.compare_params"](input_params, existing_obj)
        log.info(
            "Zabbix Value map: input params: {%s",
            str(json.dumps(input_params, indent=4)),
        )
        log.info(
            "Zabbix Value map: Object comparison result. Differences: %s",
            str(diff_params),
        )

        if diff_params:
            diff_params[zabbix_id_mapper["valuemap"]] = existing_obj[
                zabbix_id_mapper["valuemap"]
            ]
            log.info(
                "Zabbix Value map: update params: %s",
                str(json.dumps(diff_params, indent=4)),
            )

            if dry_run:
                ret["result"] = True
                ret["comment"] = 'Zabbix Value map "{}" would be fixed.'.format(name)
                ret["changes"] = {
                    name: {
                        "old": (
                            'Zabbix Value map "{}" differs '
                            "in following parameters: {}".format(name, diff_params)
                        ),
                        "new": (
                            'Zabbix Value map "{}" would correspond to definition.'.format(
                                name
                            )
                        ),
                    }
                }
            else:
                valuemap_update = __salt__["zabbix.run_query"](
                    "valuemap.update", diff_params, **kwargs
                )
                log.info(
                    "Zabbix Value map: valuemap.update result: %s",
                    str(valuemap_update),
                )
                if valuemap_update:
                    ret["result"] = True
                    ret["comment"] = 'Zabbix Value map "{}" updated.'.format(name)
                    ret["changes"] = {
                        name: {
                            "old": (
                                'Zabbix Value map "{}" differed '
                                "in following parameters: {}".format(name, diff_params)
                            ),
                            "new": 'Zabbix Value map "{}" fixed.'.format(name),
                        }
                    }

        else:
            ret["result"] = True
            ret[
                "comment"
            ] = 'Zabbix Value map "{}" already exists and corresponds to a definition.'.format(
                name
            )

    else:
        if dry_run:
            ret["result"] = True
            ret["comment"] = 'Zabbix Value map "{}" would be created.'.format(name)
            ret["changes"] = {
                name: {
                    "old": 'Zabbix Value map "{}" does not exist.'.format(name),
                    "new": (
                        'Zabbix Value map "{}" would be created '
                        "according definition.".format(name)
                    ),
                }
            }
        else:
            # ACTION.CREATE
            valuemap_create = __salt__["zabbix.run_query"](
                "valuemap.create", input_params, **kwargs
            )
            log.info(
                "Zabbix Value map: valuemap.create result: %s",
                str(valuemap_create),
            )

            if valuemap_create:
                ret["result"] = True
                ret["comment"] = 'Zabbix Value map "{}" created.'.format(name)
                ret["changes"] = {
                    name: {
                        "old": 'Zabbix Value map "{}" did not exist.'.format(name),
                        "new": (
                            'Zabbix Value map "{}" created according definition.'.format(
                                name
                            )
                        ),
                    }
                }

    return ret


def absent(name, **kwargs):
    """
    Makes the Zabbix Value map to be absent (either does not exist or delete it).

    :param name: Zabbix Value map name
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        zabbix-valuemap-absent:
            zabbix_valuemap.absent:
                - name: Value map name
    """
    dry_run = __opts__["test"]
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    try:
        object_id = __salt__["zabbix.get_object_id_by_params"](
            "valuemap", {"filter": {"name": name}}, **kwargs
        )
    except SaltException:
        object_id = False

    if not object_id:
        ret["result"] = True
        ret["comment"] = 'Zabbix Value map "{}" does not exist.'.format(name)
    else:
        if dry_run:
            ret["result"] = True
            ret["comment"] = 'Zabbix Value map "{}" would be deleted.'.format(name)
            ret["changes"] = {
                name: {
                    "old": 'Zabbix Value map "{}" exists.'.format(name),
                    "new": 'Zabbix Value map "{}" would be deleted.'.format(name),
                }
            }
        else:
            valuemap_delete = __salt__["zabbix.run_query"](
                "valuemap.delete", [object_id], **kwargs
            )

            if valuemap_delete:
                ret["result"] = True
                ret["comment"] = 'Zabbix Value map "{}" deleted.'.format(name)
                ret["changes"] = {
                    name: {
                        "old": 'Zabbix Value map "{}" existed.'.format(name),
                        "new": 'Zabbix Value map "{}" deleted.'.format(name),
                    }
                }

    return ret



"""
Tests for runner_returns
"""

import errno
import os
import socket
import tempfile

import pytest

import salt.payload
import salt.utils.args
import salt.utils.files
import salt.utils.jid
import salt.utils.yaml
from tests.support.case import ShellCase
from tests.support.runtests import RUNTIME_VARS


@pytest.mark.windows_whitelisted
class RunnerReturnsTest(ShellCase):
    """
    Test the "runner_returns" feature
    """

    def setUp(self):
        """
        Create the temp file and master.d directory
        """
        self.job_dir = os.path.join(self.master_opts["cachedir"], "jobs")
        self.hash_type = self.master_opts["hash_type"]
        self.master_d_dir = os.path.join(self.config_dir, "master.d")
        try:
            os.makedirs(self.master_d_dir)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise

        self.conf = tempfile.NamedTemporaryFile(
            mode="w",
            suffix=".conf",
            dir=self.master_d_dir,
            delete=False,
        )

    def tearDown(self):
        """
        Close the tempfile.NamedTemporaryFile object, cleaning it up
        """
        try:
            self.conf.close()
        except OSError:
            pass
        salt.utils.files.rm_rf(self.master_d_dir)
        # Force a reload of the configuration now that our temp config file has
        # been removed.
        self.run_run_plus("test.arg")

    @staticmethod
    def clean_return(data):
        """
        Remove kwargs and timestamp (things that are variable) so we have a
        stable value to assert
        """
        # Remove pub_kwargs
        data["fun_args"][1] = salt.utils.args.clean_kwargs(**data["fun_args"][1])
        data["return"]["kwargs"] = salt.utils.args.clean_kwargs(
            **data["return"]["kwargs"]
        )

        # Pop off dynamic keys in the return schema that are impossible to test.
        # Do not supply the default arguments because we want to know if we are
        # missing some aspect of the schema.
        data.pop("_stamp")

    def write_conf(self, data):
        """
        Dump the config dict to the conf file
        """
        self.conf.write(salt.utils.yaml.safe_dump(data, default_flow_style=False))
        self.conf.flush()
        self.conf.close()

    @pytest.mark.slow_test
    def test_runner_returns_disabled(self):
        """
        Test with runner_returns disabled
        """
        self.write_conf({"runner_returns": False})
        ret = self.run_run_plus("test.arg", "foo", bar="hello world!")

        jid = ret.get("jid")
        if jid is None:
            raise Exception("jid missing from run_run_plus output")

        serialized_return = os.path.join(
            salt.utils.jid.jid_dir(jid, self.job_dir, self.hash_type),
            "master",
            "return.p",
        )
        self.assertFalse(os.path.isfile(serialized_return))

    @pytest.mark.slow_test
    def test_runner_returns_enabled(self):
        """
        Test with runner_returns enabled
        """
        self.write_conf({"runner_returns": True})
        ret = self.run_run_plus("test.arg", "foo", bar="hello world!")

        jid = ret.get("jid")
        if jid is None:
            raise Exception("jid missing from run_run_plus output")

        serialized_return = os.path.join(
            salt.utils.jid.jid_dir(jid, self.job_dir, self.hash_type),
            "master",
            "return.p",
        )
        with salt.utils.files.fopen(serialized_return, "rb") as fp_:
            deserialized = salt.payload.loads(fp_.read(), encoding="utf-8")

        self.clean_return(deserialized["return"])

        # Now we have something sane we can reliably compare in an assert.
        if "SUDO_USER" in os.environ:
            user = "sudo_{}".format(os.environ["SUDO_USER"])
        else:
            user = RUNTIME_VARS.RUNNING_TESTS_USER
        if salt.utils.platform.is_windows():
            user = "sudo_{}\\{}".format(socket.gethostname(), user)
        self.assertEqual(
            deserialized,
            {
                "return": {
                    "fun": "runner.test.arg",
                    "fun_args": ["foo", {"bar": "hello world!"}],
                    "jid": jid,
                    "return": {"args": ["foo"], "kwargs": {"bar": "hello world!"}},
                    "success": True,
                    "user": user,
                }
            },
        )



"""
    tests.unit.utils.cache_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Test the salt cache objects
"""

import logging
import pathlib
import time

import pytest

import salt.loader
import salt.payload
import salt.utils.cache as cache
import salt.utils.data
import salt.utils.files
from tests.support.mock import patch


def test_sanity():
    """
    Make sure you can instantiate etc.
    """
    cd = cache.CacheDict(5)
    assert isinstance(cd, cache.CacheDict)

    # do some tests to make sure it looks like a dict
    assert "foo" not in cd
    cd["foo"] = "bar"
    assert cd["foo"] == "bar"
    del cd["foo"]
    assert "foo" not in cd


def test_ttl():
    cd = cache.CacheDict(0.1)
    cd["foo"] = "bar"
    assert "foo" in cd
    assert cd["foo"] == "bar"
    time.sleep(0.2)
    assert "foo" not in cd

    # make sure that a get would get a regular old key error
    with pytest.raises(KeyError):
        cd["foo"]  # pylint: disable=pointless-statement


@pytest.fixture
def cache_dir(minion_opts):
    return pathlib.Path(minion_opts["cachedir"])


def test_smoke_context(minion_opts):
    """
    Smoke test the context cache
    """
    context_cache = cache.ContextCache(minion_opts, "cache_test")

    data = {"a": "b"}
    context_cache.cache_context(data.copy())

    ret = context_cache.get_cache_context()

    assert ret == data


@pytest.fixture
def cache_mod_name():
    return "cache_mod"


@pytest.fixture
def cache_mods_path(tmp_path, cache_mod_name):
    _cache_mods_path = tmp_path / "cache_mods"
    mod_contents = """
    import salt.utils.cache

    def __virtual__():
        return True

    @salt.utils.cache.context_cache
    def test_context_module():
        if "called" in __context__:
            __context__["called"] += 1
        else:
            __context__["called"] = 0
        return __context__.value()

    @salt.utils.cache.context_cache
    def test_compare_context():
        return __context__.value()
    """
    with pytest.helpers.temp_file(
        cache_mod_name + ".py", mod_contents, _cache_mods_path
    ):
        yield _cache_mods_path


def test_context_wrapper(minion_opts, cache_mods_path):
    """
    Test to ensure that a module which decorates itself
    with a context cache can store and retrieve its contextual
    data
    """

    loader = salt.loader.LazyLoader(
        [str(cache_mods_path)],
        tag="rawmodule",
        virtual_enable=False,
        opts=minion_opts,
    )

    cache_test_func = loader["cache_mod.test_context_module"]

    assert cache_test_func()["called"] == 0
    assert cache_test_func()["called"] == 1


def test_set_cache(minion_opts, cache_mods_path, cache_mod_name, cache_dir):
    """
    Tests to ensure the cache is written correctly
    """

    context = {"c": "d"}
    loader = salt.loader.LazyLoader(
        [str(cache_mods_path)],
        tag="rawmodule",
        virtual_enable=False,
        opts=minion_opts,
        pack={"__context__": context, "__opts__": minion_opts},
    )

    cache_test_func = loader["cache_mod.test_context_module"]

    # Call the function to trigger the context cache
    assert cache_test_func()["called"] == 0
    assert cache_test_func()["called"] == 1
    assert cache_test_func()["called"] == 2

    cache_file_name = "salt.loaded.ext.rawmodule.{}.p".format(cache_mod_name)

    cached_file = cache_dir / "context" / cache_file_name
    assert cached_file.exists()

    # Test manual de-serialize
    target_cache_data = salt.utils.data.decode(
        salt.payload.loads(cached_file.read_bytes())
    )
    assert target_cache_data == dict(context, called=1)

    # Test cache de-serialize
    cc = cache.ContextCache(
        minion_opts, "salt.loaded.ext.rawmodule.{}".format(cache_mod_name)
    )
    retrieved_cache = cc.get_cache_context()
    assert retrieved_cache == dict(context, called=1)


def test_refill_cache(minion_opts, cache_mods_path):
    """
    Tests to ensure that the context cache can rehydrate a wrapped function
    """
    context = {"c": "d"}
    loader = salt.loader.LazyLoader(
        [str(cache_mods_path)],
        tag="rawmodule",
        virtual_enable=False,
        opts=minion_opts,
        pack={"__context__": context, "__opts__": minion_opts},
    )

    cache_test_func = loader["cache_mod.test_compare_context"]
    # First populate the cache
    ret = cache_test_func()
    assert ret == context

    # Then try to rehydrate a func
    context_copy = context.copy()
    context.clear()

    # Compare to the context before it was emptied
    ret = cache_test_func()
    assert ret == context_copy


def test_everything(cache_dir):
    """
    Make sure you can instantiate, add, update, remove, expire
    """
    path = str(cache_dir / "minion")

    # test instantiation
    cd = cache.CacheDisk(0.3, path)
    assert isinstance(cd, cache.CacheDisk)

    # test to make sure it looks like a dict
    assert "foo" not in cd
    cd["foo"] = "bar"
    assert "foo" in cd
    assert cd["foo"] == "bar"
    del cd["foo"]
    assert "foo" not in cd

    # test persistence
    cd["foo"] = "bar"
    cd2 = cache.CacheDisk(0.3, path)
    assert "foo" in cd2
    assert cd2["foo"] == "bar"

    # test ttl
    time.sleep(0.5)
    assert "foo" not in cd
    assert "foo" not in cd2


@pytest.mark.parametrize(
    "data",
    [
        b"PK\x03\x04\n\x00\x00\x00\x00\x00\xb6B\x05S\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06\x00\x1c\x00test2/",
        b"\xc3\x83\xc2\xa6\xc3\x83\xc2\xb8\xc3\x83\xc2\xa5",
    ],
)
def test_unicode_error(cache_dir, data, caplog):
    """
    Test when the data in the cache raises a UnicodeDecodeError
    we do not raise an error.
    """
    path = cache_dir / "minion"
    path.touch()
    cache_data = {
        "CacheDisk_data": {
            b"poc-minion": {
                None: {
                    b"secrets": {
                        b"itsasecret": data,
                        b"CacheDisk_cachetime": {b"poc-minion": 1649339137.1236317},
                    }
                }
            }
        }
    }
    with patch.object(
        salt.utils.msgpack, "load", return_value=cache_data
    ), caplog.at_level(logging.DEBUG):
        cd = cache.CacheDisk(0.3, str(path))
        # this test used to rely on msgpack throwing errors if attempt to read an empty file
        # code now checks if file empty and returns, so we should never attempt msgpack load
        assert cd._dict == {}
        assert not (
            f"Error reading cache file at '{path}': Unpack failed: incomplete input"
            in caplog.messages
        )


def test_cache_corruption(cache_dir):
    """
    Tests if the CacheDisk can survive a corrupted cache file.
    """
    # Write valid cache file
    cache_file = cache_dir / "minion"
    cd = cache.CacheDisk(0.3, str(cache_file))
    cd["test-key"] = "test-value"
    del cd

    # Add random string to the data to make the msgpack structure un-decodable
    with cache_file.open("a") as f:
        f.write("I am data that should corrupt the msgpack file")

    # Reopen cache, try to fetch key
    cd = cache.CacheDisk(0.3, str(cache_file))
    # If the cache is unreadable, we want it to act like an empty cache (as
    # if the file did not exist in the first place), and should raise a KeyError
    with pytest.raises(KeyError):
        assert cd["test-key"]



"""
Testing ini_manage exec module.
"""

import os
import tempfile

import salt.modules.ini_manage as ini
import salt.utils.files
import salt.utils.stringutils
from tests.support.unit import TestCase


class IniManageTestCase(TestCase):
    """
    Testing ini_manage exec module.
    """

    TEST_FILE_CONTENT = [
        "# Comment on the first line",
        "",
        "# First main option",
        "option1=main1",
        "",
        "# Second main option",
        "option2=main2",
        "",
        "",
        "[main]",
        "# Another comment",
        "test1=value 1",
        "",
        "test2=value 2",
        "",
        "[SectionB]",
        "test1=value 1B",
        "",
        "# Blank line should be above",
        "test3 = value 3B",
        "",
        "[SectionC]",
        "# The following option is empty",
        "empty_option=",
    ]

    maxDiff = None

    def _setUp(self, linesep):
        self.tfile = tempfile.NamedTemporaryFile(delete=False, mode="w+b")
        self.tfile.write(
            salt.utils.stringutils.to_bytes(linesep.join(self.TEST_FILE_CONTENT))
        )
        self.tfile.close()

    def setUp(self):
        self._setUp(os.linesep)

    def tearDown(self):
        os.remove(self.tfile.name)

    def test_get_option(self):
        """
        Test get_option method.
        """
        self.assertEqual(ini.get_option(self.tfile.name, "main", "test1"), "value 1")
        self.assertEqual(ini.get_option(self.tfile.name, "main", "test2"), "value 2")
        self.assertEqual(
            ini.get_option(self.tfile.name, "SectionB", "test1"), "value 1B"
        )
        self.assertEqual(
            ini.get_option(self.tfile.name, "SectionB", "test3"), "value 3B"
        )
        self.assertEqual(
            ini.get_option(self.tfile.name, "SectionC", "empty_option"), ""
        )

    def test_get_section(self):
        """
        Test get_section method.
        """
        self.assertEqual(
            ini.get_section(self.tfile.name, "SectionB"),
            {"test1": "value 1B", "test3": "value 3B"},
        )

    def test_remove_option(self):
        """
        Test remove_option method.
        """
        self.assertEqual(
            ini.remove_option(self.tfile.name, "SectionB", "test1"), "value 1B"
        )
        self.assertIsNone(ini.get_option(self.tfile.name, "SectionB", "test1"))

    def test_remove_section(self):
        """
        Test remove_section method.
        """
        self.assertEqual(
            ini.remove_section(self.tfile.name, "SectionB"),
            {"test1": "value 1B", "test3": "value 3B"},
        )
        self.assertEqual(ini.get_section(self.tfile.name, "SectionB"), {})

    def test_get_ini(self):
        """
        Test get_ini method.
        """
        self.assertEqual(
            dict(ini.get_ini(self.tfile.name)),
            {
                "SectionC": {"empty_option": ""},
                "SectionB": {"test1": "value 1B", "test3": "value 3B"},
                "main": {"test1": "value 1", "test2": "value 2"},
                "option2": "main2",
                "option1": "main1",
            },
        )

    def test_set_option(self):
        """
        Test set_option method.
        """
        result = ini.set_option(
            self.tfile.name,
            {
                "SectionB": {
                    "test3": "new value 3B",
                    "test_set_option": "test_set_value",
                },
                "SectionD": {"test_set_option2": "test_set_value1"},
            },
        )
        self.assertEqual(
            result,
            {
                "SectionB": {
                    "test3": {"after": "new value 3B", "before": "value 3B"},
                    "test_set_option": {"after": "test_set_value", "before": None},
                },
                "SectionD": {
                    "after": {"test_set_option2": "test_set_value1"},
                    "before": None,
                },
            },
        )
        # Check existing option updated
        self.assertEqual(
            ini.get_option(self.tfile.name, "SectionB", "test3"), "new value 3B"
        )
        # Check new section and option added
        self.assertEqual(
            ini.get_option(self.tfile.name, "SectionD", "test_set_option2"),
            "test_set_value1",
        )

    def test_empty_value(self):
        """
        Test empty value preserved after edit
        """
        ini.set_option(self.tfile.name, {"SectionB": {"test3": "new value 3B"}})
        with salt.utils.files.fopen(self.tfile.name, "r") as fp_:
            file_content = salt.utils.stringutils.to_unicode(fp_.read())
        expected = "{0}{1}{0}".format(os.linesep, "empty_option = ")
        self.assertIn(expected, file_content, "empty_option was not preserved")

    def test_empty_lines(self):
        """
        Test empty lines preserved after edit
        """
        ini.set_option(self.tfile.name, {"SectionB": {"test3": "new value 3B"}})
        expected = os.linesep.join(
            [
                "# Comment on the first line",
                "",
                "# First main option",
                "option1 = main1",
                "",
                "# Second main option",
                "option2 = main2",
                "",
                "[main]",
                "# Another comment",
                "test1 = value 1",
                "",
                "test2 = value 2",
                "",
                "[SectionB]",
                "test1 = value 1B",
                "",
                "# Blank line should be above",
                "test3 = new value 3B",
                "",
                "[SectionC]",
                "# The following option is empty",
                "empty_option = ",
                "",
            ]
        )
        with salt.utils.files.fopen(self.tfile.name, "r") as fp_:
            file_content = salt.utils.stringutils.to_unicode(fp_.read())
        self.assertEqual(expected, file_content)

    def test_empty_lines_multiple_edits(self):
        """
        Test empty lines preserved after multiple edits
        """
        ini.set_option(
            self.tfile.name,
            {"SectionB": {"test3": "this value will be edited two times"}},
        )
        self.test_empty_lines()

    def test_newline_characters(self):
        """
        Test newline characters
        """
        for c in ["\n", "\r", "\r\n"]:
            for test in [
                self.test_get_option,
                self.test_get_section,
                self.test_remove_option,
                self.test_remove_section,
                self.test_get_ini,
                self.test_set_option,
                self.test_empty_value,
                self.test_empty_lines,
                self.test_empty_lines_multiple_edits,
            ]:
                self.tearDown()
                self._setUp(c)
                test()



import os
import random
import sys
import tempfile
from contextlib import contextmanager

import pytest

import salt.utils.path
import salt.utils.platform
import salt.utils.user
from tests.support.case import ModuleCase
from tests.support.helpers import SKIP_INITIAL_PHOTONOS_FAILURES, dedent
from tests.support.runtests import RUNTIME_VARS

AVAILABLE_PYTHON_EXECUTABLE = salt.utils.path.which_bin(
    ["python", "python2", "python2.6", "python2.7"]
)


@pytest.mark.windows_whitelisted
class CMDModuleTest(ModuleCase):
    """
    Validate the cmd module
    """

    def setUp(self):
        self.runas_usr = "nobody"
        if salt.utils.platform.is_darwin():
            self.runas_usr = "macsalttest"

    @contextmanager
    def _ensure_user_exists(self, name):
        if name in self.run_function("user.info", [name]).values():
            # User already exists; don't touch
            yield
        else:
            # Need to create user for test
            self.run_function("user.add", [name])
            try:
                yield
            finally:
                self.run_function("user.delete", [name], remove=True)

    @pytest.mark.slow_test
    @pytest.mark.skip_on_windows
    def test_run(self):
        """
        cmd.run
        """
        shell = os.environ.get("SHELL")
        if shell is None:
            # Failed to get the SHELL var, don't run
            self.skipTest("Unable to get the SHELL environment variable")

        self.assertTrue(self.run_function("cmd.run", ["echo $SHELL"]))
        self.assertEqual(
            self.run_function(
                "cmd.run", ["echo $SHELL", "shell={}".format(shell)], python_shell=True
            ).rstrip(),
            shell,
        )
        self.assertEqual(
            self.run_function("cmd.run", ["ls / | grep etc"], python_shell=True), "etc"
        )
        self.assertEqual(
            self.run_function(
                "cmd.run",
                ['echo {{grains.id}} | awk "{print $1}"'],
                template="jinja",
                python_shell=True,
            ),
            "minion",
        )
        self.assertEqual(
            self.run_function(
                "cmd.run", ["grep f"], stdin="one\ntwo\nthree\nfour\nfive\n"
            ),
            "four\nfive",
        )
        self.assertEqual(
            self.run_function(
                "cmd.run", ['echo "a=b" | sed -e s/=/:/g'], python_shell=True
            ),
            "a:b",
        )

    @pytest.mark.slow_test
    def test_stdout(self):
        """
        cmd.run_stdout
        """
        self.assertEqual(
            self.run_function("cmd.run_stdout", ['echo "cheese"']).rstrip(),
            "cheese" if not salt.utils.platform.is_windows() else '"cheese"',
        )

    @pytest.mark.slow_test
    def test_stderr(self):
        """
        cmd.run_stderr
        """
        if sys.platform.startswith(("freebsd", "openbsd")):
            shell = "/bin/sh"
        else:
            shell = "/bin/bash"

        self.assertEqual(
            self.run_function(
                "cmd.run_stderr",
                ['echo "cheese" 1>&2', "shell={}".format(shell)],
                python_shell=True,
            ).rstrip(),
            "cheese" if not salt.utils.platform.is_windows() else '"cheese"',
        )

    @pytest.mark.slow_test
    def test_run_all(self):
        """
        cmd.run_all
        """
        if sys.platform.startswith(("freebsd", "openbsd")):
            shell = "/bin/sh"
        else:
            shell = "/bin/bash"

        ret = self.run_function(
            "cmd.run_all",
            ['echo "cheese" 1>&2', "shell={}".format(shell)],
            python_shell=True,
        )
        self.assertTrue("pid" in ret)
        self.assertTrue("retcode" in ret)
        self.assertTrue("stdout" in ret)
        self.assertTrue("stderr" in ret)
        self.assertTrue(isinstance(ret.get("pid"), int))
        self.assertTrue(isinstance(ret.get("retcode"), int))
        self.assertTrue(isinstance(ret.get("stdout"), str))
        self.assertTrue(isinstance(ret.get("stderr"), str))
        self.assertEqual(
            ret.get("stderr").rstrip(),
            "cheese" if not salt.utils.platform.is_windows() else '"cheese"',
        )

    @pytest.mark.slow_test
    def test_retcode(self):
        """
        cmd.retcode
        """
        self.assertEqual(
            self.run_function("cmd.retcode", ["exit 0"], python_shell=True), 0
        )
        self.assertEqual(
            self.run_function("cmd.retcode", ["exit 1"], python_shell=True), 1
        )

    @pytest.mark.slow_test
    def test_run_all_with_success_retcodes(self):
        """
        cmd.run with success_retcodes
        """
        ret = self.run_function(
            "cmd.run_all", ["exit 42"], success_retcodes=[42], python_shell=True
        )

        self.assertTrue("retcode" in ret)
        self.assertEqual(ret.get("retcode"), 0)

    @pytest.mark.slow_test
    def test_retcode_with_success_retcodes(self):
        """
        cmd.run with success_retcodes
        """
        ret = self.run_function(
            "cmd.retcode", ["exit 42"], success_retcodes=[42], python_shell=True
        )

        self.assertEqual(ret, 0)

    @pytest.mark.slow_test
    def test_run_all_with_success_stderr(self):
        """
        cmd.run with success_retcodes
        """
        random_file = "{}{}{}".format(
            RUNTIME_VARS.TMP_ROOT_DIR, os.path.sep, random.random()
        )

        if salt.utils.platform.is_windows():
            func = "type"
            expected_stderr = "cannot find the file specified"
        else:
            func = "cat"
            expected_stderr = "No such file or directory"
        ret = self.run_function(
            "cmd.run_all",
            ["{} {}".format(func, random_file)],
            success_stderr=[expected_stderr],
            python_shell=True,
        )

        self.assertTrue("retcode" in ret)
        self.assertEqual(ret.get("retcode"), 0)

    @pytest.mark.slow_test
    def test_blacklist_glob(self):
        """
        cmd_blacklist_glob
        """
        self.assertEqual(
            self.run_function("cmd.run", ["bad_command --foo"]).rstrip(),
            'ERROR: The shell command "bad_command --foo" is not permitted',
        )

    @pytest.mark.slow_test
    def test_script(self):
        """
        cmd.script
        """
        args = "saltines crackers biscuits=yes"
        script = "salt://script.py"
        ret = self.run_function("cmd.script", [script, args], saltenv="base")
        self.assertEqual(ret["stdout"], args)

    @pytest.mark.slow_test
    def test_script_query_string(self):
        """
        cmd.script
        """
        args = "saltines crackers biscuits=yes"
        script = "salt://script.py?saltenv=base"
        ret = self.run_function("cmd.script", [script, args], saltenv="base")
        self.assertEqual(ret["stdout"], args)

    @pytest.mark.slow_test
    def test_script_retcode(self):
        """
        cmd.script_retcode
        """
        script = "salt://script.py"
        ret = self.run_function("cmd.script_retcode", [script], saltenv="base")
        self.assertEqual(ret, 0)

    @pytest.mark.slow_test
    def test_script_cwd(self):
        """
        cmd.script with cwd
        """
        tmp_cwd = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        args = "saltines crackers biscuits=yes"
        script = "salt://script.py"
        ret = self.run_function(
            "cmd.script", [script, args], cwd=tmp_cwd, saltenv="base"
        )
        self.assertEqual(ret["stdout"], args)

    @pytest.mark.slow_test
    def test_script_cwd_with_space(self):
        """
        cmd.script with cwd
        """
        tmp_cwd = "{}{}test 2".format(
            tempfile.mkdtemp(dir=RUNTIME_VARS.TMP), os.path.sep
        )
        os.mkdir(tmp_cwd)

        args = "saltines crackers biscuits=yes"
        script = "salt://script.py"
        ret = self.run_function(
            "cmd.script", [script, args], cwd=tmp_cwd, saltenv="base"
        )
        self.assertEqual(ret["stdout"], args)

    @pytest.mark.destructive_test
    def test_tty(self):
        """
        cmd.tty
        """
        for tty in ("tty0", "pts3"):
            if os.path.exists(os.path.join("/dev", tty)):
                ret = self.run_function("cmd.tty", [tty, "apply salt liberally"])
                self.assertTrue("Success" in ret)

    @pytest.mark.skip_on_windows
    @pytest.mark.skip_if_binaries_missing("which")
    def test_which(self):
        """
        cmd.which
        """
        cmd_which = self.run_function("cmd.which", ["cat"])
        self.assertIsInstance(cmd_which, str)
        cmd_run = self.run_function("cmd.run", ["which cat"])
        self.assertIsInstance(cmd_run, str)
        self.assertEqual(cmd_which.rstrip(), cmd_run.rstrip())

    @pytest.mark.skip_on_windows
    @pytest.mark.skip_if_binaries_missing("which")
    def test_which_bin(self):
        """
        cmd.which_bin
        """
        cmds = ["pip3", "pip2", "pip", "pip-python"]
        ret = self.run_function("cmd.which_bin", [cmds])
        self.assertTrue(os.path.split(ret)[1] in cmds)

    @pytest.mark.slow_test
    def test_has_exec(self):
        """
        cmd.has_exec
        """
        self.assertTrue(
            self.run_function("cmd.has_exec", [AVAILABLE_PYTHON_EXECUTABLE])
        )
        self.assertFalse(
            self.run_function("cmd.has_exec", ["alllfsdfnwieulrrh9123857ygf"])
        )

    @pytest.mark.slow_test
    def test_exec_code(self):
        """
        cmd.exec_code
        """
        code = dedent(
            """
                   import sys
                   sys.stdout.write('cheese')
               """
        )
        self.assertEqual(
            self.run_function(
                "cmd.exec_code", [AVAILABLE_PYTHON_EXECUTABLE, code]
            ).rstrip(),
            "cheese",
        )

    @pytest.mark.slow_test
    def test_exec_code_with_single_arg(self):
        """
        cmd.exec_code
        """
        code = dedent(
            """
                   import sys
                   sys.stdout.write(sys.argv[1])
               """
        )
        arg = "cheese"
        self.assertEqual(
            self.run_function(
                "cmd.exec_code", [AVAILABLE_PYTHON_EXECUTABLE, code], args=arg
            ).rstrip(),
            arg,
        )

    @pytest.mark.slow_test
    def test_exec_code_with_multiple_args(self):
        """
        cmd.exec_code
        """
        code = dedent(
            """
                   import sys
                   sys.stdout.write(sys.argv[1])
               """
        )
        arg = "cheese"
        self.assertEqual(
            self.run_function(
                "cmd.exec_code", [AVAILABLE_PYTHON_EXECUTABLE, code], args=[arg, "test"]
            ).rstrip(),
            arg,
        )

    @pytest.mark.slow_test
    def test_quotes(self):
        """
        cmd.run with quoted command
        """
        cmd = """echo 'SELECT * FROM foo WHERE bar="baz"' """
        expected_result = 'SELECT * FROM foo WHERE bar="baz"'
        if salt.utils.platform.is_windows():
            expected_result = "'SELECT * FROM foo WHERE bar=\"baz\"'"
        result = self.run_function("cmd.run_stdout", [cmd]).strip()
        self.assertEqual(result, expected_result)

    @pytest.mark.skip_if_not_root
    @pytest.mark.skip_on_windows(reason="Skip on Windows, requires password")
    def test_quotes_runas(self):
        """
        cmd.run with quoted command
        """
        cmd = """echo 'SELECT * FROM foo WHERE bar="baz"' """
        expected_result = 'SELECT * FROM foo WHERE bar="baz"'
        result = self.run_function(
            "cmd.run_all", [cmd], runas=RUNTIME_VARS.RUNNING_TESTS_USER
        )
        errmsg = "The command returned: {}".format(result)
        self.assertEqual(result["retcode"], 0, errmsg)
        self.assertEqual(result["stdout"], expected_result, errmsg)

    @pytest.mark.destructive_test
    @pytest.mark.skip_if_not_root
    @pytest.mark.skip_on_windows(reason="Skip on Windows, uses unix commands")
    @pytest.mark.slow_test
    def test_avoid_injecting_shell_code_as_root(self):
        """
        cmd.run should execute the whole command as the "runas" user, not
        running substitutions as root.
        """
        cmd = "echo $(id -u)"

        root_id = self.run_function("cmd.run_stdout", [cmd])
        runas_root_id = self.run_function(
            "cmd.run_stdout", [cmd], runas=RUNTIME_VARS.RUNNING_TESTS_USER
        )
        with self._ensure_user_exists(self.runas_usr):
            user_id = self.run_function("cmd.run_stdout", [cmd], runas=self.runas_usr)

        self.assertNotEqual(user_id, root_id)
        self.assertNotEqual(user_id, runas_root_id)
        self.assertEqual(root_id, runas_root_id)

    @pytest.mark.destructive_test
    @pytest.mark.skip_if_not_root
    @pytest.mark.skip_on_windows(reason="Skip on Windows, uses unix commands")
    @pytest.mark.slow_test
    def test_cwd_runas(self):
        """
        cmd.run should be able to change working directory correctly, whether
        or not runas is in use.
        """
        cmd = "pwd"
        tmp_cwd = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        os.chmod(tmp_cwd, 0o711)

        cwd_normal = self.run_function("cmd.run_stdout", [cmd], cwd=tmp_cwd).rstrip(
            "\n"
        )
        self.assertEqual(tmp_cwd, cwd_normal)

        with self._ensure_user_exists(self.runas_usr):
            cwd_runas = self.run_function(
                "cmd.run_stdout", [cmd], cwd=tmp_cwd, runas=self.runas_usr
            ).rstrip("\n")
        self.assertEqual(tmp_cwd, cwd_runas)

    @pytest.mark.destructive_test
    @pytest.mark.skip_if_not_root
    @pytest.mark.skip_unless_on_darwin(reason="Applicable to MacOS only")
    @pytest.mark.slow_test
    def test_runas_env(self):
        """
        cmd.run should be able to change working directory correctly, whether
        or not runas is in use.
        """
        with self._ensure_user_exists(self.runas_usr):
            user_path = self.run_function(
                "cmd.run_stdout", ['printf %s "$PATH"'], runas=self.runas_usr
            )
        # XXX: Not sure of a better way. Environment starts out with
        # /bin:/usr/bin and should be populated by path helper and the bash
        # profile.
        self.assertNotEqual("/bin:/usr/bin", user_path)

    @pytest.mark.destructive_test
    @pytest.mark.skip_if_not_root
    @pytest.mark.skip_unless_on_darwin(reason="Applicable to MacOS only")
    @pytest.mark.slow_test
    def test_runas_complex_command_bad_cwd(self):
        """
        cmd.run should not accidentally run parts of a complex command when
        given a cwd which cannot be used by the user the command is run as.

        Due to the need to use `su -l` to login to another user on MacOS, we
        cannot cd into directories that the target user themselves does not
        have execute permission for. To an extent, this test is testing that
        buggy behaviour, but its purpose is to ensure that the greater bug of
        running commands after failing to cd does not occur.
        """
        tmp_cwd = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        os.chmod(tmp_cwd, 0o700)

        with self._ensure_user_exists(self.runas_usr):
            cmd_result = self.run_function(
                "cmd.run_all",
                ['pwd; pwd; : $(echo "You have failed the test" >&2)'],
                cwd=tmp_cwd,
                runas=self.runas_usr,
            )

        self.assertEqual("", cmd_result["stdout"])
        self.assertNotIn("You have failed the test", cmd_result["stderr"])
        self.assertNotEqual(0, cmd_result["retcode"])

    @SKIP_INITIAL_PHOTONOS_FAILURES
    @pytest.mark.skip_on_windows
    @pytest.mark.skip_if_not_root
    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_runas(self):
        """
        Ensure that the env is the runas user's
        """
        with self._ensure_user_exists(self.runas_usr):
            out = self.run_function(
                "cmd.run", ["env"], runas=self.runas_usr
            ).splitlines()
        self.assertIn("USER={}".format(self.runas_usr), out)

    @pytest.mark.skip_if_binaries_missing("sleep", reason="sleep cmd not installed")
    def test_timeout(self):
        """
        cmd.run trigger timeout
        """
        out = self.run_function(
            "cmd.run", ["sleep 2 && echo hello"], f_timeout=1, python_shell=True
        )
        self.assertTrue("Timed out" in out)

    @pytest.mark.skip_if_binaries_missing("sleep", reason="sleep cmd not installed")
    def test_timeout_success(self):
        """
        cmd.run sufficient timeout to succeed
        """
        out = self.run_function(
            "cmd.run", ["sleep 1 && echo hello"], f_timeout=2, python_shell=True
        )
        self.assertEqual(out, "hello")

    @pytest.mark.slow_test
    def test_hide_output(self):
        """
        Test the hide_output argument
        """
        ls_command = (
            ["ls", "/"] if not salt.utils.platform.is_windows() else ["dir", "c:\\"]
        )

        error_command = ["thiscommanddoesnotexist"]

        # cmd.run
        out = self.run_function("cmd.run", ls_command, hide_output=True)
        self.assertEqual(out, "")

        # cmd.shell
        out = self.run_function("cmd.shell", ls_command, hide_output=True)
        self.assertEqual(out, "")

        # cmd.run_stdout
        out = self.run_function("cmd.run_stdout", ls_command, hide_output=True)
        self.assertEqual(out, "")

        # cmd.run_stderr
        out = self.run_function("cmd.shell", error_command, hide_output=True)
        self.assertEqual(out, "")

        # cmd.run_all (command should have produced stdout)
        out = self.run_function("cmd.run_all", ls_command, hide_output=True)
        self.assertEqual(out["stdout"], "")
        self.assertEqual(out["stderr"], "")

        # cmd.run_all (command should have produced stderr)
        out = self.run_function("cmd.run_all", error_command, hide_output=True)
        self.assertEqual(out["stdout"], "")
        self.assertEqual(out["stderr"], "")

    @pytest.mark.slow_test
    def test_cmd_run_whoami(self):
        """
        test return of whoami
        """
        if not salt.utils.platform.is_windows():
            user = RUNTIME_VARS.RUNTIME_CONFIGS["master"]["user"]
        else:
            user = salt.utils.user.get_specific_user()
        if user.startswith("sudo_"):
            user = user.replace("sudo_", "")
        cmd = self.run_function("cmd.run", ["whoami"])
        try:
            self.assertEqual(user.lower(), cmd.lower())
        except AssertionError as exc:
            if not salt.utils.platform.is_windows():
                raise exc from None
            if "\\" in user:
                user = user.split("\\")[-1]
            self.assertEqual(user.lower(), cmd.lower())

    @pytest.mark.skip_unless_on_windows(reason="Minion is not Windows")
    @pytest.mark.slow_test
    def test_windows_env_handling(self):
        """
        Ensure that nt.environ is used properly with cmd.run*
        """
        out = self.run_function(
            "cmd.run", ["set"], env={"abc": "123", "ABC": "456"}
        ).splitlines()
        self.assertIn("abc=123", out)
        self.assertIn("ABC=456", out)

    @pytest.mark.slow_test
    @pytest.mark.skip_unless_on_windows(reason="Minion is not Windows")
    def test_windows_powershell_script_args(self):
        """
        Ensure that powershell processes inline script in args
        """
        val = "i like cheese"
        args = (
            '-SecureString (ConvertTo-SecureString -String "{}" -AsPlainText -Force)'
            " -ErrorAction Stop".format(val)
        )
        script = "salt://issue-56195/test.ps1"
        ret = self.run_function(
            "cmd.script", [script], args=args, shell="powershell", saltenv="base"
        )
        self.assertEqual(ret["stdout"], val)

    @pytest.mark.slow_test
    @pytest.mark.skip_unless_on_windows(reason="Minion is not Windows")
    @pytest.mark.skip_if_binaries_missing("pwsh")
    def test_windows_powershell_script_args_pwsh(self):
        """
        Ensure that powershell processes inline script in args with powershell
        core
        """
        val = "i like cheese"
        args = (
            '-SecureString (ConvertTo-SecureString -String "{}" -AsPlainText -Force)'
            " -ErrorAction Stop".format(val)
        )
        script = "salt://issue-56195/test.ps1"
        ret = self.run_function(
            "cmd.script", [script], args=args, shell="pwsh", saltenv="base"
        )
        self.assertEqual(ret["stdout"], val)



# Python libs
import datetime
import logging
import time

import pytest

# Salt libs
from salt.beacons import telegram_bot_msg

# Salt testing libs
from tests.support.mock import MagicMock, patch

# Third-party libs
try:
    import telegram

    HAS_TELEGRAM = True
except ImportError:
    HAS_TELEGRAM = False


log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skipif(HAS_TELEGRAM is False, reason="telegram is not available"),
]


@pytest.fixture
def configure_loader_modules():
    return {telegram_bot_msg: {}}


def test_validate_empty_config(*args, **kwargs):
    ret = telegram_bot_msg.validate(None)
    assert ret == (False, "Configuration for telegram_bot_msg beacon must be a list.")


def test_validate_missing_accept_from_config(*args, **kwargs):
    ret = telegram_bot_msg.validate([{"token": "bcd"}])
    assert ret == (
        False,
        "Not all required configuration for telegram_bot_msg are set.",
    )


def test_validate_missing_token_config(*args, **kwargs):
    ret = telegram_bot_msg.validate([{"accept_from": []}])
    assert ret == (
        False,
        "Not all required configuration for telegram_bot_msg are set.",
    )


def test_validate_config_not_list_in_accept_from(*args, **kwargs):
    ret = telegram_bot_msg.validate([{"token": "bcd", "accept_from": {"nodict": "1"}}])
    assert ret == (
        False,
        "Configuration for telegram_bot_msg, "
        "accept_from must be a list of "
        "usernames.",
    )


def test_validate_valid_config(*args, **kwargs):
    ret = telegram_bot_msg.validate([{"token": "bcd", "accept_from": ["username"]}])
    assert ret == (True, "Valid beacon configuration.")


def test_call_no_updates():
    with patch("salt.beacons.telegram_bot_msg.telegram") as telegram_api:
        token = "abc"
        config = [{"token": token, "accept_from": ["tester"]}]
        inst = MagicMock(name="telegram.Bot()")
        telegram_api.Bot = MagicMock(name="telegram", return_value=inst)
        inst.get_updates.return_value = []

        ret = telegram_bot_msg.validate(config)
        assert ret == (True, "Valid beacon configuration.")

        ret = telegram_bot_msg.beacon(config)
        telegram_api.Bot.assert_called_once_with(token)
        assert ret == []


def test_call_telegram_return_no_updates_for_user():
    with patch("salt.beacons.telegram_bot_msg.telegram") as telegram_api:
        token = "abc"
        username = "tester"
        config = [{"token": token, "accept_from": [username]}]
        inst = MagicMock(name="telegram.Bot()")
        telegram_api.Bot = MagicMock(name="telegram", return_value=inst)

        log.debug("telegram %s", telegram)
        username = "different_user"
        user = telegram.user.User(id=1, first_name="", username=username, is_bot=True)
        chat = telegram.chat.Chat(1, "private", username=username)
        date = time.mktime(datetime.datetime(2016, 12, 18, 0, 0).timetuple())
        message = telegram.message.Message(
            message_id=1, from_user=user, date=date, chat=chat
        )
        update = telegram.update.Update(update_id=1, message=message)

        inst.get_updates.return_value = [update]

        ret = telegram_bot_msg.validate(config)
        assert ret == (True, "Valid beacon configuration.")

        ret = telegram_bot_msg.beacon(config)
        telegram_api.Bot.assert_called_once_with(token)
        assert ret == []


def test_call_telegram_returning_updates():
    with patch("salt.beacons.telegram_bot_msg.telegram") as telegram_api:
        token = "abc"
        username = "tester"
        config = [{"token": token, "accept_from": [username]}]
        inst = MagicMock(name="telegram.Bot()")
        telegram_api.Bot = MagicMock(name="telegram", return_value=inst)

        user = telegram.User(id=1, first_name="", username=username, is_bot=True)
        chat = telegram.Chat(1, "private", username=username)
        date = time.mktime(datetime.datetime(2016, 12, 18, 0, 0).timetuple())
        message = telegram.Message(message_id=1, from_user=user, date=date, chat=chat)
        update = telegram.update.Update(update_id=1, message=message)

        inst.get_updates.return_value = [update]

        ret = telegram_bot_msg.validate(config)
        assert ret == (True, "Valid beacon configuration.")

        ret = telegram_bot_msg.beacon(config)
        telegram_api.Bot.assert_called_once_with(token)
        assert ret
        assert ret[0]["msgs"][0] == message.to_dict()



"""
    :codeauthor: :email:`Amel Ajdinovic <amel@stackpointcloud.com>`
"""
import pytest

from tests.integration.cloud.helpers.cloud_test_base import TIMEOUT, CloudTest

try:
    from oneandone.client import OneAndOneService  # pylint: disable=unused-import

    HAS_ONEANDONE = True
except ImportError:
    HAS_ONEANDONE = False


@pytest.mark.skipif(HAS_ONEANDONE is False, reason="salt-cloud requires >= 1and1 1.2.0")
class OneAndOneTest(CloudTest):
    """
    Integration tests for the 1and1 cloud provider
    """

    PROVIDER = "oneandone"
    REQUIRED_PROVIDER_CONFIG_ITEMS = ("api_token",)

    def test_list_images(self):
        """
        Tests the return of running the --list-images command for 1and1
        """
        image_list = self.run_cloud("--list-images {}".format(self.PROVIDER_NAME))
        self.assertIn("coreOSimage", [i.strip() for i in image_list])

    def test_instance(self):
        """
        Test creating an instance on 1and1
        """
        # check if instance with salt installed returned
        ret_str = self.run_cloud(
            "-p oneandone-test {}".format(self.instance_name), timeout=TIMEOUT
        )
        self.assertInstanceExists(ret_str)

        self.assertDestroyInstance()



"""
:codeauthor: Shane Lee <slee@saltstack.com>
"""
import pytest

import salt.modules.cmdmod as cmdmod
import salt.modules.win_file as win_file
import salt.modules.win_lgpo as win_lgpo
import salt.utils.win_dacl as win_dacl
import salt.utils.win_lgpo_auditpol as win_lgpo_auditpol
import salt.utils.win_reg as win_reg

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(tmp_path):
    cachedir = tmp_path / "__test_admx_policy_cache_dir"
    cachedir.mkdir(parents=True, exist_ok=True)
    return {
        win_lgpo: {
            "__salt__": {
                "cmd.run": cmdmod.run,
                "file.file_exists": win_file.file_exists,
                "file.makedirs": win_file.makedirs_,
                "file.remove": win_file.remove,
                "file.write": win_file.write,
            },
            "__opts__": {
                "cachedir": str(cachedir),
            },
            "__utils__": {
                "auditpol.get_auditpol_dump": win_lgpo_auditpol.get_auditpol_dump,
                "reg.read_value": win_reg.read_value,
            },
        },
        win_file: {
            "__utils__": {
                "dacl.set_perms": win_dacl.set_perms,
            },
        },
    }


def test__getAdmlPresentationRefId():
    resources = win_lgpo._get_policy_resources(language="en-US")
    ref_id = "LetAppsAccessAccountInfo_Enum"
    expected = "Default for all apps"
    result = win_lgpo._getAdmlPresentationRefId(resources, ref_id)
    assert result == expected


def test__getAdmlPresentationRefId_result_text_is_none():
    resources = win_lgpo._get_policy_resources(language="en-US")
    ref_id = "LetAppsAccessAccountInfo_UserInControlOfTheseApps_List"
    expected = "Put user in control of these specific apps (use Package Family Names)"
    result = win_lgpo._getAdmlPresentationRefId(resources, ref_id)
    assert result == expected



import functools
import hashlib
import http.server
import multiprocessing
import os
import random
import shutil
import socket
import sys
from contextlib import closing

import pytest

import salt.utils.files


class RequestHandler(http.server.SimpleHTTPRequestHandler):
    """
    Modified request handler class
    """

    def __init__(self, *args, directory=None, **kwargs):
        if directory is None:
            directory = os.getcwd()
        self.directory = directory
        if sys.version_info.minor < 7:
            super().__init__(*args, **kwargs)
        else:
            super().__init__(*args, directory=directory, **kwargs)

    def do_GET(self):
        """
        GET request handling
        """
        none_match = self.headers.get("If-None-Match")
        status_code = 200
        try:
            # Retrieve the local file from the web root to serve to clients
            with salt.utils.files.fopen(
                os.path.join(self.directory, self.path[1:])
            ) as reqfp:
                return_text = reqfp.read().encode("utf-8")
                # We're using this checksum as the etag to show file changes
                checksum = hashlib.md5(return_text).hexdigest()
                if none_match == checksum:
                    # Status code 304 Not Modified is returned if the file is unchanged
                    status_code = 304
        except:  # pylint: disable=bare-except
            # Something went wrong. We didn't find the requested file
            status_code = 404
            return_text = None
            checksum = None

        self.send_response(status_code)

        # Return the Etag header if we have the checksum
        if checksum:
            # IMPORTANT: This introduces randomness into the tests. The Etag header key
            # will be converted to lowercase in the code... but if someone breaks that,
            # it'll rear it's head here as random failures that are hard to reproduce.
            # Any alternatives seem overly complex. So... don't break the case insensitivity
            # in the code.
            possible_etags = ["Etag", "ETag"]
            self.send_header(random.choice(possible_etags), checksum)
            self.end_headers()

        # Return file content
        if return_text:
            self.wfile.write(return_text)


def serve(port=8000, directory=None):
    """
    Function to serve a directory via http.server
    """
    handler = functools.partial(RequestHandler, directory=directory)
    s = http.server.HTTPServer(("127.0.0.1", port), handler)
    s.serve_forever()


@pytest.fixture(scope="module")
def free_port():
    """
    Utility fixture to grab a free port for the web server
    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


@pytest.fixture(autouse=True, scope="module")
def server(free_port, web_root):
    """
    Web server fixture
    """
    p = multiprocessing.Process(target=serve, args=(free_port, web_root))
    p.start()
    yield
    p.terminate()
    p.join()


@pytest.fixture(scope="module")
def web_root(tmp_path_factory):
    """
    Temporary directory fixture for the web server root
    """
    _web_root = tmp_path_factory.mktemp("web_root")
    try:
        yield str(_web_root)
    finally:
        shutil.rmtree(str(_web_root), ignore_errors=True)


@pytest.mark.slow_test
def test_file_managed_web_source_etag_operation(
    states, free_port, web_root, minion_opts
):
    """
    This functional test checks the operation of the use_etag parameter to the
    file.managed state. There are four (4) invocations of file.managed with a
    web source, but only three (3) will trigger a call to the web server as
    shown below and in comments within.

        127.0.0.1 - - [08/Jan/2022 00:53:11] "GET /foo.txt HTTP/1.1" 200 -
        127.0.0.1 - - [08/Jan/2022 00:53:11] "GET /foo.txt HTTP/1.1" 304 -
        127.0.0.1 - - [08/Jan/2022 00:53:12] "GET /foo.txt HTTP/1.1" 200 -

    Checks are documented in the comments.
    """
    # Create file in the web root directory to serve
    states.file.managed(
        name=os.path.join(web_root, "foo.txt"), contents="this is my file"
    )

    # File should not be cached yet
    cached_file = os.path.join(
        minion_opts["cachedir"],
        "extrn_files",
        "base",
        "localhost{free_port}".format(free_port=free_port),
        "foo.txt",
    )
    cached_etag = cached_file + ".etag"
    assert not os.path.exists(cached_file)
    assert not os.path.exists(cached_etag)

    # Pull the file from the web server
    #     Web server returns 200 status code with content:
    #     127.0.0.1 - - [08/Jan/2022 00:53:11] "GET /foo.txt HTTP/1.1" 200 -
    states.file.managed(
        name=os.path.join(web_root, "bar.txt"),
        source="http://localhost:{free_port}/foo.txt".format(free_port=free_port),
        use_etag=True,
    )

    # Now the file is cached
    assert os.path.exists(cached_file)
    assert os.path.exists(cached_etag)

    # Store the original modified time of the cached file
    cached_file_mtime = os.path.getmtime(cached_file)

    # Pull the file again. Etag hasn't changed. No download occurs.
    #     Web server returns 304 status code and no content:
    #     127.0.0.1 - - [08/Jan/2022 00:53:11] "GET /foo.txt HTTP/1.1" 304 -
    states.file.managed(
        name=os.path.join(web_root, "bar.txt"),
        source="http://localhost:{free_port}/foo.txt".format(free_port=free_port),
        use_etag=True,
    )

    # Check that the modified time of the cached file hasn't changed
    assert cached_file_mtime == os.path.getmtime(cached_file)

    # Change file in the web root directory
    states.file.managed(
        name=os.path.join(web_root, "foo.txt"), contents="this is my changed file"
    )

    # Don't use Etag. Cached file is there, Salt won't try to download.
    #     No call to the web server will be made.
    states.file.managed(
        name=os.path.join(web_root, "bar.txt"),
        source="http://localhost:{free_port}/foo.txt".format(free_port=free_port),
        use_etag=False,
    )

    # Check that the modified time of the cached file hasn't changed
    assert cached_file_mtime == os.path.getmtime(cached_file)

    # Now use Etag again. Cached file changes
    #     Web server returns 200 status code with content
    #     127.0.0.1 - - [08/Jan/2022 00:53:12] "GET /foo.txt HTTP/1.1" 200 -
    states.file.managed(
        name=os.path.join(web_root, "bar.txt"),
        source="http://localhost:{free_port}/foo.txt".format(free_port=free_port),
        use_etag=True,
    )

    # The modified time of the cached file now changes
    assert cached_file_mtime != os.path.getmtime(cached_file)


def test_file_symlink_replace_existing_link(states, tmp_path):
    # symlink name and target for state
    name = tmp_path / "foo"
    target = tmp_path / "baz"

    # create existing symlink to replace
    old_target = tmp_path / "bar"
    name.symlink_to(old_target)

    ret = states.file.symlink(
        name=str(name),
        target=str(target),
    )

    assert ret.filtered == {
        "name": str(name),
        "changes": {"new": str(name)},
        "comment": f"Created new symlink {str(name)} -> {str(target)}",
        "result": True,
    }



import salt.pillar.pillar_ldap as pillar_ldap
import salt.utils.stringutils


def test__config_returns_str():
    conf = {"foo": "bar"}
    assert pillar_ldap._config("foo", conf) == salt.utils.stringutils.to_str("bar")


def test__conf_defaults_to_none():
    conf = {"foo": "bar"}
    assert pillar_ldap._config("bang", conf) is None


def test__conf_returns_str_from_unicode_default():
    conf = {"foo": "bar"}
    default = salt.utils.stringutils.to_unicode("bam")
    assert pillar_ldap._config("bang", conf, default) == salt.utils.stringutils.to_str(
        "bam"
    )



"""
Subversion SCM
"""

import re

import salt.utils.args
import salt.utils.path
from salt.exceptions import CommandExecutionError

_INI_RE = re.compile(r"^([^:]+):\s+(\S.*)$", re.M)


def __virtual__():
    """
    Only load if svn is installed
    """
    if salt.utils.path.which("svn") is None:
        return (False, "The svn execution module cannot be loaded: svn unavailable.")
    else:
        return True


def _run_svn(cmd, cwd, user, username, password, opts, **kwargs):
    """
    Execute svn
    return the output of the command

    cmd
        The command to run.

    cwd
        The path to the Subversion repository

    user
        Run svn as a user other than what the minion runs as

    username
        Connect to the Subversion server as another user

    password
        Connect to the Subversion server with this password

        .. versionadded:: 0.17.0

    opts
        Any additional options to add to the command line

    kwargs
        Additional options to pass to the run-cmd
    """
    cmd = ["svn", "--non-interactive", cmd]

    options = list(opts)
    if username:
        options.extend(["--username", username])
    if password:
        options.extend(["--password", password])
    cmd.extend(options)

    result = __salt__["cmd.run_all"](
        cmd, python_shell=False, cwd=cwd, runas=user, **kwargs
    )

    retcode = result["retcode"]

    if retcode == 0:
        return result["stdout"]
    raise CommandExecutionError(result["stderr"] + "\n\n" + " ".join(cmd))


def info(cwd, targets=None, user=None, username=None, password=None, fmt="str"):
    """
    Display the Subversion information from the checkout.

    cwd
        The path to the Subversion repository

    targets : None
        files, directories, and URLs to pass to the command as arguments
        svn uses '.' by default

    user : None
        Run svn as a user other than what the minion runs as

    username : None
        Connect to the Subversion server as another user

    password : None
        Connect to the Subversion server with this password

        .. versionadded:: 0.17.0

    fmt : str
        How to fmt the output from info.
        (str, xml, list, dict)

    CLI Example:

    .. code-block:: bash

        salt '*' svn.info /path/to/svn/repo
    """
    opts = list()
    if fmt == "xml":
        opts.append("--xml")
    if targets:
        opts += salt.utils.args.shlex_split(targets)
    infos = _run_svn("info", cwd, user, username, password, opts)

    if fmt in ("str", "xml"):
        return infos

    info_list = []
    for infosplit in infos.split("\n\n"):
        info_list.append(_INI_RE.findall(infosplit))

    if fmt == "list":
        return info_list
    if fmt == "dict":
        return [dict(tmp) for tmp in info_list]


def checkout(cwd, remote, target=None, user=None, username=None, password=None, *opts):
    """
    Download a working copy of the remote Subversion repository
    directory or file

    cwd
        The path to the Subversion repository

    remote : None
        URL to checkout

    target : None
        The name to give the file or directory working copy
        Default: svn uses the remote basename

    user : None
        Run svn as a user other than what the minion runs as

    username : None
        Connect to the Subversion server as another user

    password : None
        Connect to the Subversion server with this password

        .. versionadded:: 0.17.0

    CLI Example:

    .. code-block:: bash

        salt '*' svn.checkout /path/to/repo svn://remote/repo
    """
    opts += (remote,)
    if target:
        opts += (target,)
    return _run_svn("checkout", cwd, user, username, password, opts)


def switch(cwd, remote, target=None, user=None, username=None, password=None, *opts):
    """
    .. versionadded:: 2014.1.0

    Switch a working copy of a remote Subversion repository
    directory

    cwd
        The path to the Subversion repository

    remote : None
        URL to switch

    target : None
        The name to give the file or directory working copy
        Default: svn uses the remote basename

    user : None
        Run svn as a user other than what the minion runs as

    username : None
        Connect to the Subversion server as another user

    password : None
        Connect to the Subversion server with this password

    CLI Example:

    .. code-block:: bash

        salt '*' svn.switch /path/to/repo svn://remote/repo
    """
    opts += (remote,)
    if target:
        opts += (target,)
    return _run_svn("switch", cwd, user, username, password, opts)


def update(cwd, targets=None, user=None, username=None, password=None, *opts):
    """
    Update the current directory, files, or directories from
    the remote Subversion repository

    cwd
        The path to the Subversion repository

    targets : None
        files and directories to pass to the command as arguments
        Default: svn uses '.'

    user : None
        Run svn as a user other than what the minion runs as

    password : None
        Connect to the Subversion server with this password

        .. versionadded:: 0.17.0

    username : None
        Connect to the Subversion server as another user

    CLI Example:

    .. code-block:: bash

        salt '*' svn.update /path/to/repo
    """
    if targets:
        opts += tuple(salt.utils.args.shlex_split(targets))
    return _run_svn("update", cwd, user, username, password, opts)


def diff(cwd, targets=None, user=None, username=None, password=None, *opts):
    """
    Return the diff of the current directory, files, or directories from
    the remote Subversion repository

    cwd
        The path to the Subversion repository

    targets : None
        files and directories to pass to the command as arguments
        Default: svn uses '.'

    user : None
        Run svn as a user other than what the minion runs as

    username : None
        Connect to the Subversion server as another user

    password : None
        Connect to the Subversion server with this password

        .. versionadded:: 0.17.0

    CLI Example:

    .. code-block:: bash

        salt '*' svn.diff /path/to/repo
    """
    if targets:
        opts += tuple(salt.utils.args.shlex_split(targets))
    return _run_svn("diff", cwd, user, username, password, opts)


def commit(cwd, targets=None, msg=None, user=None, username=None, password=None, *opts):
    """
    Commit the current directory, files, or directories to
    the remote Subversion repository

    cwd
        The path to the Subversion repository

    targets : None
        files and directories to pass to the command as arguments
        Default: svn uses '.'

    msg : None
        Message to attach to the commit log

    user : None
        Run svn as a user other than what the minion runs as

    username : None
        Connect to the Subversion server as another user

    password : None
        Connect to the Subversion server with this password

        .. versionadded:: 0.17.0

    CLI Example:

    .. code-block:: bash

        salt '*' svn.commit /path/to/repo
    """
    if msg:
        opts += ("-m", msg)
    if targets:
        opts += tuple(salt.utils.args.shlex_split(targets))
    return _run_svn("commit", cwd, user, username, password, opts)


def add(cwd, targets, user=None, username=None, password=None, *opts):
    """
    Add files to be tracked by the Subversion working-copy checkout

    cwd
        The path to the Subversion repository

    targets : None
        files and directories to pass to the command as arguments

    user : None
        Run svn as a user other than what the minion runs as

    username : None
        Connect to the Subversion server as another user

    password : None
        Connect to the Subversion server with this password

        .. versionadded:: 0.17.0

    CLI Example:

    .. code-block:: bash

        salt '*' svn.add /path/to/repo /path/to/new/file
    """
    if targets:
        opts += tuple(salt.utils.args.shlex_split(targets))
    return _run_svn("add", cwd, user, username, password, opts)


def remove(cwd, targets, msg=None, user=None, username=None, password=None, *opts):
    """
    Remove files and directories from the Subversion repository

    cwd
        The path to the Subversion repository

    targets : None
        files, directories, and URLs to pass to the command as arguments

    msg : None
        Message to attach to the commit log

    user : None
        Run svn as a user other than what the minion runs as

    username : None
        Connect to the Subversion server as another user

    password : None
        Connect to the Subversion server with this password

        .. versionadded:: 0.17.0

    CLI Example:

    .. code-block:: bash

        salt '*' svn.remove /path/to/repo /path/to/repo/remove
    """
    if msg:
        opts += ("-m", msg)
    if targets:
        opts += tuple(salt.utils.args.shlex_split(targets))
    return _run_svn("remove", cwd, user, username, password, opts)


def status(cwd, targets=None, user=None, username=None, password=None, *opts):
    """
    Display the status of the current directory, files, or
    directories in the Subversion repository

    cwd
        The path to the Subversion repository

    targets : None
        files, directories, and URLs to pass to the command as arguments
        Default: svn uses '.'

    user : None
        Run svn as a user other than what the minion runs as

    username : None
        Connect to the Subversion server as another user

    password : None
        Connect to the Subversion server with this password

        .. versionadded:: 0.17.0

    CLI Example:

    .. code-block:: bash

        salt '*' svn.status /path/to/repo
    """
    if targets:
        opts += tuple(salt.utils.args.shlex_split(targets))
    return _run_svn("status", cwd, user, username, password, opts)


def export(
    cwd,
    remote,
    target=None,
    user=None,
    username=None,
    password=None,
    revision="HEAD",
    *opts
):
    """
    Create an unversioned copy of a tree.

    cwd
        The path to the Subversion repository

    remote : None
        URL and path to file or directory checkout

    target : None
        The name to give the file or directory working copy
        Default: svn uses the remote basename

    user : None
        Run svn as a user other than what the minion runs as

    username : None
        Connect to the Subversion server as another user

    password : None
        Connect to the Subversion server with this password

        .. versionadded:: 0.17.0

    CLI Example:

    .. code-block:: bash

        salt '*' svn.export /path/to/repo svn://remote/repo
    """
    opts += (remote,)
    if target:
        opts += (target,)
    revision_args = "-r"
    opts += (
        revision_args,
        str(revision),
    )
    return _run_svn("export", cwd, user, username, password, opts)



"""
Detect MDADM RAIDs
"""

import logging

import salt.utils.files

log = logging.getLogger(__name__)


def mdadm():
    """
    Return list of mdadm devices
    """
    devices = set()
    try:
        with salt.utils.files.fopen("/proc/mdstat", "r") as mdstat:
            for line in mdstat:
                line = salt.utils.stringutils.to_unicode(line)
                if line.startswith("Personalities : "):
                    continue
                if line.startswith("unused devices:"):
                    continue
                if " : " in line:
                    devices.add(line.split(" : ")[0])
    except OSError:
        return {}

    devices = sorted(devices)
    if devices:
        log.trace("mdadm devices detected: %s", ", ".join(devices))

    return {"mdadm": devices}



"""
    Test cases for salt.modules.helm
"""


import pytest

import salt.states.helm as helm
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {helm: {}}


def test_repo_managed_import_failed_repo_manage():
    ret = {
        "name": "state_id",
        "changes": {},
        "result": False,
        "comment": "'helm.repo_manage' modules not available on this minion.",
    }
    assert helm.repo_managed("state_id") == ret


def test_repo_managed_import_failed_repo_update():
    mock_helm_modules = {"helm.repo_manage": MagicMock(return_value=True)}
    with patch.dict(helm.__salt__, mock_helm_modules):
        ret = {
            "name": "state_id",
            "changes": {},
            "result": False,
            "comment": "'helm.repo_update' modules not available on this minion.",
        }
        assert helm.repo_managed("state_id") == ret


def test_repo_managed_is_testing():
    mock_helm_modules = {
        "helm.repo_manage": MagicMock(return_value=True),
        "helm.repo_update": MagicMock(return_value=True),
    }
    with patch.dict(helm.__salt__, mock_helm_modules):
        mock__opts__ = {"test": MagicMock(return_value=True)}
        with patch.dict(helm.__opts__, mock__opts__):
            ret = {
                "name": "state_id",
                "result": None,
                "comment": "Helm repo would have been managed.",
                "changes": {},
            }
            assert helm.repo_managed("state_id") == ret


def test_repo_managed_success():
    result_changes = {"added": True, "removed": True, "failed": False}
    mock_helm_modules = {
        "helm.repo_manage": MagicMock(return_value=result_changes),
        "helm.repo_update": MagicMock(return_value=True),
    }
    with patch.dict(helm.__salt__, mock_helm_modules):
        ret = {
            "name": "state_id",
            "result": True,
            "comment": "Repositories were added or removed.",
            "changes": result_changes,
        }
        assert helm.repo_managed("state_id") == ret


def test_repo_managed_success_with_update():
    result_changes = {"added": True, "removed": True, "failed": False}
    mock_helm_modules = {
        "helm.repo_manage": MagicMock(return_value=result_changes),
        "helm.repo_update": MagicMock(return_value=True),
    }
    result_wanted = result_changes
    result_wanted.update({"repo_update": True})
    with patch.dict(helm.__salt__, mock_helm_modules):
        ret = {
            "name": "state_id",
            "result": True,
            "comment": "Repositories were added or removed.",
            "changes": result_wanted,
        }
        assert helm.repo_managed("state_id") == ret


def test_repo_managed_failed():
    result_changes = {"added": True, "removed": True, "failed": True}
    mock_helm_modules = {
        "helm.repo_manage": MagicMock(return_value=result_changes),
        "helm.repo_update": MagicMock(return_value=True),
    }
    with patch.dict(helm.__salt__, mock_helm_modules):
        ret = {
            "name": "state_id",
            "result": False,
            "comment": "Failed to add or remove some repositories.",
            "changes": result_changes,
        }
        assert helm.repo_managed("state_id") == ret


def test_repo_updated_import_failed():
    ret = {
        "name": "state_id",
        "changes": {},
        "result": False,
        "comment": "'helm.repo_update' modules not available on this minion.",
    }
    assert helm.repo_updated("state_id") == ret


def test_repo_updated_is_testing():
    mock_helm_modules = {"helm.repo_update": MagicMock(return_value=True)}
    with patch.dict(helm.__salt__, mock_helm_modules):
        mock__opts__ = {"test": MagicMock(return_value=True)}
        with patch.dict(helm.__opts__, mock__opts__):
            ret = {
                "name": "state_id",
                "result": None,
                "comment": "Helm repo would have been updated.",
                "changes": {},
            }
            assert helm.repo_updated("state_id") == ret


def test_repo_updated_success():
    mock_helm_modules = {"helm.repo_update": MagicMock(return_value=True)}
    with patch.dict(helm.__salt__, mock_helm_modules):
        ret = {
            "name": "state_id",
            "result": True,
            "comment": "Helm repo is updated.",
            "changes": {},
        }
        assert helm.repo_updated("state_id") == ret


def test_repo_updated_failed():
    mock_helm_modules = {"helm.repo_update": MagicMock(return_value=False)}
    with patch.dict(helm.__salt__, mock_helm_modules):
        ret = {
            "name": "state_id",
            "result": False,
            "comment": "Failed to sync some repositories.",
            "changes": False,
        }
        assert helm.repo_updated("state_id") == ret


def test_release_present_import_failed_helm_status():
    ret = {
        "name": "state_id",
        "changes": {},
        "result": False,
        "comment": "'helm.status' modules not available on this minion.",
    }
    assert helm.release_present("state_id", "mychart") == ret


def test_release_present_import_failed_helm_install():
    mock_helm_modules = {"helm.status": MagicMock(return_value=True)}
    with patch.dict(helm.__salt__, mock_helm_modules):
        ret = {
            "name": "state_id",
            "changes": {},
            "result": False,
            "comment": "'helm.install' modules not available on this minion.",
        }
        assert helm.release_present("state_id", "mychart") == ret


def test_release_present_import_failed_helm_upgrade():
    mock_helm_modules = {
        "helm.status": MagicMock(return_value=True),
        "helm.install": MagicMock(return_value=True),
    }
    with patch.dict(helm.__salt__, mock_helm_modules):
        ret = {
            "name": "state_id",
            "changes": {},
            "result": False,
            "comment": "'helm.upgrade' modules not available on this minion.",
        }
        assert helm.release_present("state_id", "mychart") == ret


def test_release_present_is_testing():
    mock_helm_modules = {
        "helm.status": MagicMock(return_value=True),
        "helm.install": MagicMock(return_value=True),
        "helm.upgrade": MagicMock(return_value=True),
    }
    with patch.dict(helm.__salt__, mock_helm_modules):
        mock__opts__ = {"test": MagicMock(return_value=True)}
        with patch.dict(helm.__opts__, mock__opts__):
            ret = {
                "name": "state_id",
                "result": None,
                "comment": "Helm release would have been installed or updated.",
                "changes": {},
            }
            assert helm.release_present("state_id", "mychart") == ret


def test_release_absent_import_failed_helm_uninstall():
    ret = {
        "name": "state_id",
        "changes": {},
        "result": False,
        "comment": "'helm.uninstall' modules not available on this minion.",
    }
    assert helm.release_absent("state_id") == ret


def test_release_absent_import_failed_helm_status():
    mock_helm_modules = {"helm.uninstall": MagicMock(return_value=True)}
    with patch.dict(helm.__salt__, mock_helm_modules):
        ret = {
            "name": "state_id",
            "changes": {},
            "result": False,
            "comment": "'helm.status' modules not available on this minion.",
        }
        assert helm.release_absent("state_id") == ret


def test_release_absent_is_testing():
    mock_helm_modules = {
        "helm.status": MagicMock(return_value=True),
        "helm.uninstall": MagicMock(return_value=True),
    }
    with patch.dict(helm.__salt__, mock_helm_modules):
        mock__opts__ = {"test": MagicMock(return_value=True)}
        with patch.dict(helm.__opts__, mock__opts__):
            ret = {
                "name": "state_id",
                "result": None,
                "comment": "Helm release would have been uninstalled.",
                "changes": {},
            }
            assert helm.release_absent("state_id") == ret


def test_release_absent_success():
    mock_helm_modules = {
        "helm.status": MagicMock(return_value={}),
        "helm.uninstall": MagicMock(return_value=True),
    }
    with patch.dict(helm.__salt__, mock_helm_modules):
        ret = {
            "name": "state_id",
            "result": True,
            "comment": "Helm release state_id is absent.",
            "changes": {"absent": "state_id"},
        }
        assert helm.release_absent("state_id") == ret


def test_release_absent_error():
    mock_helm_modules = {
        "helm.status": MagicMock(return_value={}),
        "helm.uninstall": MagicMock(return_value="error"),
    }
    with patch.dict(helm.__salt__, mock_helm_modules):
        ret = {
            "name": "state_id",
            "result": False,
            "comment": "error",
            "changes": {},
        }
        assert helm.release_absent("state_id") == ret



import pytest

import salt.modules.openbsdrcctl_service as openbsdrcctl
from tests.support.mock import MagicMock, patch


@pytest.fixture(autouse=True)
def rcctl():
    cmd = "rcctl"
    with patch.object(openbsdrcctl, "_cmd", return_value=cmd):
        yield cmd


@pytest.fixture
def retcode_mock():
    return MagicMock()


@pytest.fixture
def configure_loader_modules(retcode_mock):
    return {
        openbsdrcctl: {
            "__salt__": {"cmd.retcode": retcode_mock},
        },
    }


def test_available(retcode_mock, rcctl):
    retcode_mock.return_value = 0
    assert openbsdrcctl.available("test") is True
    retcode_mock.assert_called_with("{} get test".format(rcctl), ignore_retcode=True)
    retcode_mock.return_value = 2
    assert openbsdrcctl.available("test") is False
    retcode_mock.assert_called_with("{} get test".format(rcctl), ignore_retcode=True)


def test_status(retcode_mock, rcctl):
    retcode_mock.return_value = 0
    assert openbsdrcctl.status("test") is True
    retcode_mock.assert_called_with("{} check test".format(rcctl), ignore_retcode=True)
    retcode_mock.return_value = 2
    assert openbsdrcctl.status("test") is False
    retcode_mock.assert_called_with("{} check test".format(rcctl), ignore_retcode=True)


def test_disabled(retcode_mock, rcctl):
    retcode_mock.return_value = 0
    assert openbsdrcctl.disabled("test") is False
    retcode_mock.assert_called_with(
        "{} get test status".format(rcctl), ignore_retcode=True
    )
    retcode_mock.return_value = 2
    assert openbsdrcctl.disabled("test") is True
    retcode_mock.assert_called_with(
        "{} get test status".format(rcctl), ignore_retcode=True
    )


def test_enabled(retcode_mock, rcctl):
    retcode_mock.return_value = 0
    flags_return = {"flag1": "value1"}
    stdout_mock = MagicMock(return_value=flags_return)
    salt_mock = {
        "cmd.run_stdout": stdout_mock,
        "config.option": MagicMock(),
    }
    with patch.dict(openbsdrcctl.__salt__, salt_mock):
        assert openbsdrcctl.enabled("test", flags=flags_return) is True
        retcode_mock.assert_called_with(
            "{} get test status".format(rcctl), ignore_retcode=True
        )
        retcode_mock.return_value = 2
        stdout_mock.reset_mock()
        assert openbsdrcctl.enabled("test") is False
        retcode_mock.assert_called_with(
            "{} get test status".format(rcctl), ignore_retcode=True
        )



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""
import os

import pytest
import yaml

import salt.utils.cloud
import salt.utils.files
import salt.utils.yaml
from tests.integration.cloud.helpers.cloud_test_base import CloudTest
from tests.support import win_installer
from tests.support.runtests import RUNTIME_VARS

HAS_WINRM = salt.utils.cloud.HAS_WINRM and salt.utils.cloud.HAS_SMB
# THis test needs a longer timeout than other cloud tests
TIMEOUT = 1200


class EC2Test(CloudTest):
    """
    Integration tests for the EC2 cloud provider in Salt-Cloud
    """

    PROVIDER = "ec2"
    REQUIRED_PROVIDER_CONFIG_ITEMS = ("id", "key", "keyname", "private_key", "location")

    @staticmethod
    def __fetch_installer():
        # Determine the downloaded installer name by searching the files
        # directory for the first file that looks like an installer.
        for path, dirs, files in os.walk(RUNTIME_VARS.FILES):
            for file in files:
                if file.startswith(win_installer.PREFIX):
                    return file

        # If the installer wasn't found in the previous steps, download the latest Windows installer executable
        name = win_installer.latest_installer_name()
        path = os.path.join(RUNTIME_VARS.FILES, name)
        with salt.utils.files.fopen(path, "wb") as fp:
            win_installer.download_and_verify(fp, name)
        return name

    @property
    def installer(self):
        """
        Make sure the testing environment has a Windows installer executable.
        """
        if not hasattr(self, "_installer"):
            self._installer = self.__fetch_installer()
        return self._installer

    def setUp(self):
        """
        Sets up the test requirements
        """
        group_or_subnet = self.provider_config.get("securitygroup")
        if not group_or_subnet:
            group_or_subnet = self.provider_config.get("subnetid")

        if not group_or_subnet:
            self.skipTest(
                "securitygroup or subnetid missing for {} config".format(self.PROVIDER)
            )

        super().setUp()

    def override_profile_config(self, name, data):
        conf_path = os.path.join(
            RUNTIME_VARS.TMP_CONF_DIR, "cloud.profiles.d", "ec2.conf"
        )
        with salt.utils.files.fopen(conf_path, "r") as fp:
            conf = yaml.safe_load(fp)
        conf[name].update(data)
        with salt.utils.files.fopen(conf_path, "w") as fp:
            salt.utils.yaml.safe_dump(conf, fp)

    def copy_file(self, name):
        """
        Copy a file from tests/integration/files to a test's temporary
        configuration directory. The path to the file which is created will be
        returned.
        """
        src = os.path.join(RUNTIME_VARS.FILES, name)
        dst = os.path.join(RUNTIME_VARS.TMP_CONF_DIR, name)
        with salt.utils.files.fopen(src, "rb") as sfp:
            with salt.utils.files.fopen(dst, "wb") as dfp:
                dfp.write(sfp.read())
        return dst

    def _test_instance(self, profile, debug):
        """
        Tests creating and deleting an instance on EC2 (classic)
        """

        # create the instance
        cmd = ["-p", profile]
        if debug:
            cmd.extend(["-l", "debug"])
        cmd.append(self.instance_name)
        ret_val = self.run_cloud(" ".join(cmd), timeout=TIMEOUT)

        # check if instance returned with salt installed
        self.assertInstanceExists(ret_val)
        ipv6Address_present = False
        for each in ret_val:
            if "ipv6Address:" in each:
                ipv6Address_present = True
        assert ipv6Address_present

        self.assertDestroyInstance()

    def test_instance_rename(self):
        """
        Tests creating and renaming an instance on EC2 (classic)
        """
        # create the instance
        ret_val = self.run_cloud(
            "-p ec2-test {} --no-deploy".format(self.instance_name), timeout=TIMEOUT
        )
        # check if instance returned
        self.assertInstanceExists(ret_val)

        changed_name = self.instance_name + "-changed"

        rename_result = self.run_cloud(
            "-a rename {} newname={} --assume-yes".format(
                self.instance_name, changed_name
            ),
            timeout=TIMEOUT,
        )
        self.assertFalse(
            self._instance_exists(),
            "Instance wasn't renamed: |\n{}".format(rename_result),
        )
        self.assertInstanceExists(instance_name=changed_name)

        self.assertDestroyInstance(changed_name)

    def test_instance(self):
        """
        Tests creating and deleting an instance on EC2 (classic)
        """
        self._test_instance("ec2-test", debug=False)

    def test_win2012r2_psexec(self):
        """
        Tests creating and deleting a Windows 2012r2instance on EC2 using
        psexec (classic)
        """
        # TODO: psexec calls hang and the test fails by timing out. The same
        # same calls succeed when run outside of the test environment.
        # FIXME? Does this override need to be undone at the end of the test?
        self.override_profile_config(
            "ec2-win2012r2-test",
            {
                "use_winrm": False,
                "userdata_file": self.copy_file("windows-firewall-winexe.ps1"),
                "win_installer": self.copy_file(self.installer),
            },
        )
        self._test_instance("ec2-win2012r2-test", debug=True)

    @pytest.mark.skipif(
        not HAS_WINRM, reason="Skip when winrm dependencies are missing"
    )
    def test_win2012r2_winrm(self):
        """
        Tests creating and deleting a Windows 2012r2 instance on EC2 using
        winrm (classic)
        """
        self.override_profile_config(
            "ec2-win2012r2-test",
            {
                "userdata_file": self.copy_file("windows-firewall.ps1"),
                "win_installer": self.copy_file(self.installer),
                "winrm_ssl_verify": False,
                "use_winrm": True,
            },
        )
        self._test_instance("ec2-win2012r2-test", debug=True)

    def test_win2016_psexec(self):
        """
        Tests creating and deleting a Windows 2016 instance on EC2 using winrm
        (classic)
        """
        # TODO: winexe calls hang and the test fails by timing out. The
        # same calls succeed when run outside of the test environment.
        self.override_profile_config(
            "ec2-win2016-test",
            {
                "use_winrm": False,
                "userdata_file": self.copy_file("windows-firewall-winexe.ps1"),
                "win_installer": self.copy_file(self.installer),
            },
        )
        self._test_instance("ec2-win2016-test", debug=True)

    @pytest.mark.skipif(
        not HAS_WINRM, reason="Skip when winrm dependencies are missing"
    )
    def test_win2016_winrm(self):
        """
        Tests creating and deleting a Windows 2016 instance on EC2 using winrm
        (classic)
        """
        self.override_profile_config(
            "ec2-win2016-test",
            {
                "userdata_file": self.copy_file("windows-firewall.ps1"),
                "win_installer": self.copy_file(self.installer),
                "winrm_ssl_verify": False,
                "use_winrm": True,
            },
        )
        self._test_instance("ec2-win2016-test", debug=True)



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)

    tests.support.mock
    ~~~~~~~~~~~~~~~~~~

    Helper module that wraps `mock` and provides some fake objects in order to
    properly set the function/class decorators and yet skip the test case's
    execution.

    Note: mock >= 2.0.0 required since unittest.mock does not have
    MagicMock.assert_called in Python < 3.6.
"""
# pylint: disable=unused-import,function-redefined,blacklisted-module,blacklisted-external-module


import copy
import errno
import fnmatch
import sys

# By these days, we should blowup if mock is not available
import mock  # pylint: disable=blacklisted-external-import

# pylint: disable=no-name-in-module,no-member
from mock import (
    ANY,
    DEFAULT,
    FILTER_DIR,
    AsyncMock,
    MagicMock,
    Mock,
    NonCallableMagicMock,
    NonCallableMock,
    PropertyMock,
    __version__,
    call,
    create_autospec,
    patch,
    sentinel,
)

import salt.utils.stringutils

# pylint: disable=no-name-in-module,no-member


__mock_version = tuple(
    int(part) for part in mock.__version__.split(".") if part.isdigit()
)  # pylint: disable=no-member


class MockFH:
    def __init__(self, filename, read_data, *args, **kwargs):
        self.filename = filename
        self.read_data = read_data
        try:
            self.mode = args[0]
        except IndexError:
            self.mode = kwargs.get("mode", "r")
        self.binary_mode = "b" in self.mode
        self.read_mode = any(x in self.mode for x in ("r", "+"))
        self.write_mode = any(x in self.mode for x in ("w", "a", "+"))
        self.empty_string = b"" if self.binary_mode else ""
        self.call = MockCall(filename, *args, **kwargs)
        self.read_data_iter = self._iterate_read_data(read_data)
        self.read = Mock(side_effect=self._read)
        self.readlines = Mock(side_effect=self._readlines)
        self.readline = Mock(side_effect=self._readline)
        self.write = Mock(side_effect=self._write)
        self.writelines = Mock(side_effect=self._writelines)
        self.close = Mock()
        self.seek = Mock()
        self.__loc = 0
        self.__read_data_ok = False

    def _iterate_read_data(self, read_data):
        """
        Helper for mock_open:
        Retrieve lines from read_data via a generator so that separate calls to
        readline, read, and readlines are properly interleaved
        """
        # Newline will always be a bytestring on PY2 because mock_open will have
        # normalized it to one.
        newline = b"\n" if isinstance(read_data, bytes) else "\n"

        read_data = [line + newline for line in read_data.split(newline)]

        if read_data[-1] == newline:
            # If the last line ended in a newline, the list comprehension will have an
            # extra entry that's just a newline. Remove this.
            read_data = read_data[:-1]
        else:
            # If there wasn't an extra newline by itself, then the file being
            # emulated doesn't have a newline to end the last line, so remove the
            # newline that we added in the list comprehension.
            read_data[-1] = read_data[-1][:-1]

        yield from read_data

    @property
    def write_calls(self):
        """
        Return a list of all calls to the .write() mock
        """
        return [x[1][0] for x in self.write.mock_calls]

    @property
    def writelines_calls(self):
        """
        Return a list of all calls to the .writelines() mock
        """
        return [x[1][0] for x in self.writelines.mock_calls]

    def tell(self):
        return self.__loc

    def __check_read_data(self):
        if not self.__read_data_ok:
            if self.binary_mode:
                if not isinstance(self.read_data, bytes):
                    raise TypeError(
                        "{} opened in binary mode, expected read_data to be "
                        "bytes, not {}".format(
                            self.filename, type(self.read_data).__name__
                        )
                    )
            else:
                if not isinstance(self.read_data, str):
                    raise TypeError(
                        "{} opened in non-binary mode, expected read_data to "
                        "be str, not {}".format(
                            self.filename, type(self.read_data).__name__
                        )
                    )
            # No need to repeat this the next time we check
            self.__read_data_ok = True

    def _read(self, size=0):
        self.__check_read_data()
        if not self.read_mode:
            raise OSError("File not open for reading")
        if not isinstance(size, int) or size < 0:
            raise TypeError("a positive integer is required")

        joined = self.empty_string.join(self.read_data_iter)
        if not size:
            # read() called with no args, return everything
            self.__loc += len(joined)
            return joined
        else:
            # read() called with an explicit size. Return a slice matching the
            # requested size, but before doing so, reset read_data to reflect
            # what we read.
            self.read_data_iter = self._iterate_read_data(joined[size:])
            ret = joined[:size]
            self.__loc += len(ret)
            return ret

    def _readlines(self, size=None):  # pylint: disable=unused-argument
        # TODO: Implement "size" argument
        self.__check_read_data()
        if not self.read_mode:
            raise OSError("File not open for reading")
        ret = list(self.read_data_iter)
        self.__loc += sum(len(x) for x in ret)
        return ret

    def _readline(self, size=None):  # pylint: disable=unused-argument
        # TODO: Implement "size" argument
        self.__check_read_data()
        if not self.read_mode:
            raise OSError("File not open for reading")
        try:
            ret = next(self.read_data_iter)
            self.__loc += len(ret)
            return ret
        except StopIteration:
            return self.empty_string

    def __iter__(self):
        self.__check_read_data()
        if not self.read_mode:
            raise OSError("File not open for reading")
        while True:
            try:
                ret = next(self.read_data_iter)
                self.__loc += len(ret)
                yield ret
            except StopIteration:
                break

    def _write(self, content):
        if not self.write_mode:
            raise OSError("File not open for writing")
        else:
            content_type = type(content)
            if self.binary_mode and content_type is not bytes:
                raise TypeError(
                    "a bytes-like object is required, not '{}'".format(
                        content_type.__name__
                    )
                )
            elif not self.binary_mode and content_type is not str:
                raise TypeError(
                    "write() argument must be str, not {}".format(content_type.__name__)
                )

    def _writelines(self, lines):
        if not self.write_mode:
            raise OSError("File not open for writing")
        for line in lines:
            self._write(line)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # pylint: disable=unused-argument
        pass


class MockCall:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        ret = "MockCall("
        for arg in self.args:
            ret += repr(arg) + ", "
        if not self.kwargs:
            if self.args:
                # Remove trailing ', '
                ret = ret[:-2]
        else:
            for key, val in self.kwargs.items():
                ret += "{}={}".format(salt.utils.stringutils.to_str(key), repr(val))
        ret += ")"
        return ret

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return self.args == other.args and self.kwargs == other.kwargs


class MockOpen:
    r'''
    This class can be used to mock the use of ``open()``.

    ``read_data`` is a string representing the contents of the file to be read.
    By default, this is an empty string.

    Optionally, ``read_data`` can be a dictionary mapping ``fnmatch.fnmatch()``
    patterns to strings (or optionally, exceptions). This allows the mocked
    filehandle to serve content for more than one file path.

    .. code-block:: python

        data = {
            '/etc/foo.conf': textwrap.dedent("""\
                Foo
                Bar
                Baz
                """),
            '/etc/bar.conf': textwrap.dedent("""\
                A
                B
                C
                """),
        }
        with patch('salt.utils.files.fopen', mock_open(read_data=data):
            do stuff

    If the file path being opened does not match any of the glob expressions,
    an IOError will be raised to simulate the file not existing.

    Passing ``read_data`` as a string is equivalent to passing it with a glob
    expression of "*". That is to say, the below two invocations are
    equivalent:

    .. code-block:: python

        mock_open(read_data='foo\n')
        mock_open(read_data={'*': 'foo\n'})

    Instead of a string representing file contents, ``read_data`` can map to an
    exception, and that exception will be raised if a file matching that
    pattern is opened:

    .. code-block:: python

        data = {
            '/etc/*': IOError(errno.EACCES, 'Permission denied'),
            '*': 'Hello world!\n',
        }
        with patch('salt.utils.files.fopen', mock_open(read_data=data)):
            do stuff

    The above would raise an exception if any files within /etc are opened, but
    would produce a mocked filehandle if any other file is opened.

    To simulate file contents changing upon subsequent opens, the file contents
    can be a list of strings/exceptions. For example:

    .. code-block:: python

        data = {
            '/etc/foo.conf': [
                'before\n',
                'after\n',
            ],
            '/etc/bar.conf': [
                IOError(errno.ENOENT, 'No such file or directory', '/etc/bar.conf'),
                'Hey, the file exists now!',
            ],
        }
        with patch('salt.utils.files.fopen', mock_open(read_data=data):
            do stuff

    The first open of ``/etc/foo.conf`` would return "before\n" when read,
    while the second would return "after\n" when read. For ``/etc/bar.conf``,
    the first read would raise an exception, while the second would open
    successfully and read the specified string.

    Expressions will be attempted in dictionary iteration order (the exception
    being ``*`` which is tried last), so if a file path matches more than one
    fnmatch expression then the first match "wins". If your use case calls for
    overlapping expressions, then an OrderedDict can be used to ensure that the
    desired matching behavior occurs:

    .. code-block:: python

        data = OrderedDict()
        data['/etc/foo.conf'] = 'Permission granted!'
        data['/etc/*'] = IOError(errno.EACCES, 'Permission denied')
        data['*'] = '*': 'Hello world!\n'
        with patch('salt.utils.files.fopen', mock_open(read_data=data):
            do stuff

    The following attributes are tracked for the life of a mock object:

    * call_count - Tracks how many fopen calls were attempted
    * filehandles - This is a dictionary mapping filenames to lists of MockFH
      objects, representing the individual times that a given file was opened.
    '''

    def __init__(self, read_data=""):
        # If the read_data contains lists, we will be popping it. So, don't
        # modify the original value passed.
        read_data = copy.copy(read_data)

        # Normalize read_data, Python 2 filehandles should never produce unicode
        # types on read.
        if not isinstance(read_data, dict):
            read_data = {"*": read_data}

        self.read_data = read_data
        self.filehandles = {}
        self.calls = []
        self.call_count = 0

    def __call__(self, name, *args, **kwargs):
        """
        Match the file being opened to the patterns in the read_data and spawn
        a mocked filehandle with the corresponding file contents.
        """
        call = MockCall(name, *args, **kwargs)
        self.calls.append(call)
        self.call_count += 1
        for pat in self.read_data:
            if pat == "*":
                continue
            if fnmatch.fnmatch(name, pat):
                matched_pattern = pat
                break
        else:
            # No non-glob match in read_data, fall back to '*'
            matched_pattern = "*"
        try:
            matched_contents = self.read_data[matched_pattern]
            try:
                # Assuming that the value for the matching expression is a
                # list, pop the first element off of it.
                file_contents = matched_contents.pop(0)
            except AttributeError:
                # The value for the matching expression is a string (or exception)
                file_contents = matched_contents
            except IndexError:
                # We've run out of file contents, abort!
                raise RuntimeError(
                    "File matching expression '{}' opened more times than "
                    "expected".format(matched_pattern)
                )

            try:
                # Raise the exception if the matched file contents are an
                # instance of an exception class.
                raise file_contents
            except TypeError:
                # Contents were not an exception, so proceed with creating the
                # mocked filehandle.
                pass

            ret = MockFH(name, file_contents, *args, **kwargs)
            self.filehandles.setdefault(name, []).append(ret)
            return ret
        except KeyError:
            # No matching glob in read_data, treat this as a file that does
            # not exist and raise the appropriate exception.
            raise OSError(errno.ENOENT, "No such file or directory", name)

    def write_calls(self, path=None):
        """
        Returns the contents passed to all .write() calls. Use `path` to narrow
        the results to files matching a given pattern.
        """
        ret = []
        for filename, handles in self.filehandles.items():
            if path is None or fnmatch.fnmatch(filename, path):
                for fh_ in handles:
                    ret.extend(fh_.write_calls)
        return ret

    def writelines_calls(self, path=None):
        """
        Returns the contents passed to all .writelines() calls. Use `path` to
        narrow the results to files matching a given pattern.
        """
        ret = []
        for filename, handles in self.filehandles.items():
            if path is None or fnmatch.fnmatch(filename, path):
                for fh_ in handles:
                    ret.extend(fh_.writelines_calls)
        return ret


class MockTimedProc:
    """
    Class used as a stand-in for salt.utils.timed_subprocess.TimedProc
    """

    class _Process:
        """
        Used to provide a dummy "process" attribute
        """

        def __init__(self, returncode=0, pid=12345):
            self.returncode = returncode
            self.pid = pid

    def __init__(self, stdout=None, stderr=None, returncode=0, pid=12345):
        if stdout is not None and not isinstance(stdout, bytes):
            raise TypeError("Must pass stdout to MockTimedProc as bytes")
        if stderr is not None and not isinstance(stderr, bytes):
            raise TypeError("Must pass stderr to MockTimedProc as bytes")
        self._stdout = stdout
        self._stderr = stderr
        self.process = self._Process(returncode=returncode, pid=pid)

    def run(self):
        pass

    @property
    def stdout(self):
        return self._stdout

    @property
    def stderr(self):
        return self._stderr


# reimplement mock_open to support multiple filehandles
mock_open = MockOpen



"""
The `pillar_roots` wheel module is used to manage files under the pillar roots
directories on the master server.
"""

import os

import salt.utils.files
import salt.utils.path
import salt.utils.verify


def find(path, saltenv="base"):
    """
    Return a dict of the files located with the given path and environment
    """
    # Return a list of paths + text or bin
    ret = []
    if saltenv not in __opts__["pillar_roots"]:
        return ret
    for root in __opts__["pillar_roots"][saltenv]:
        full = os.path.join(root, path)
        if os.path.isfile(full):
            # Add it to the dict
            with salt.utils.files.fopen(full, "rb") as fp_:
                if salt.utils.files.is_text(fp_):
                    ret.append({full: "txt"})
                else:
                    ret.append({full: "bin"})
    return ret


def list_env(saltenv="base"):
    """
    Return all of the file paths found in an environment
    """
    ret = {}
    if saltenv not in __opts__["pillar_roots"]:
        return ret
    for f_root in __opts__["pillar_roots"][saltenv]:
        ret[f_root] = {}
        for root, dirs, files in salt.utils.path.os_walk(f_root):
            sub = ret[f_root]
            if root != f_root:
                # grab subroot ref
                sroot = root
                above = []
                # Populate the above dict
                while not os.path.samefile(sroot, f_root):
                    base = os.path.basename(sroot)
                    if base:
                        above.insert(0, base)
                    sroot = os.path.dirname(sroot)
                for aroot in above:
                    sub = sub[aroot]
            for dir_ in dirs:
                sub[dir_] = {}
            for fn_ in files:
                sub[fn_] = "f"
    return ret


def list_roots():
    """
    Return all of the files names in all available environments
    """
    ret = {}
    for saltenv in __opts__["pillar_roots"]:
        ret[saltenv] = []
        ret[saltenv].append(list_env(saltenv))
    return ret


def read(path, saltenv="base"):
    """
    Read the contents of a text file, if the file is binary then
    """
    # Return a dict of paths + content
    ret = []
    files = find(path, saltenv)
    for fn_ in files:
        full = next(iter(fn_.keys()))
        form = fn_[full]
        if form == "txt":
            with salt.utils.files.fopen(full, "rb") as fp_:
                ret.append({full: salt.utils.stringutils.to_unicode(fp_.read())})
    return ret


def write(data, path, saltenv="base", index=0):
    """
    Write the named file, by default the first file found is written, but the
    index of the file can be specified to write to a lower priority file root
    """
    if saltenv not in __opts__["pillar_roots"]:
        return "Named environment {} is not present".format(saltenv)
    if len(__opts__["pillar_roots"][saltenv]) <= index:
        return "Specified index {} in environment {} is not present".format(
            index, saltenv
        )
    if os.path.isabs(path):
        return "The path passed in {} is not relative to the environment {}".format(
            path, saltenv
        )
    roots_dir = __opts__["pillar_roots"][saltenv][index]
    dest = os.path.join(roots_dir, path)
    if not salt.utils.verify.clean_path(roots_dir, dest, subdir=True):
        return "Invalid path"
    dest = os.path.join(__opts__["pillar_roots"][saltenv][index], path)
    dest_dir = os.path.dirname(dest)
    if not os.path.isdir(dest_dir):
        os.makedirs(dest_dir)
    with salt.utils.files.fopen(dest, "w+") as fp_:
        fp_.write(salt.utils.stringutils.to_str(data))
    return "Wrote data to file {}".format(dest)
